package agent;

import geneticAlgorithm.State;
import geometry.Degree;
import geometry.Line;
import geometry.Mathematic;
import geometry.Point;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;
import java.util.Map.Entry;

import rescuecore2.messages.Command;
import rescuecore2.registry.Registry;
import rescuecore2.standard.components.StandardAgent;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import utilities.Logger;
import utilities.Timer;
import utilities.config.Config;
import worldGraph.Enterance;
import worldGraph.WorldGraph;

public abstract class Agent<E extends StandardEntity> extends StandardAgent<E> {
	public Config psdConfig = null;
	public Logger logger = null;
	protected Timer cycleTimer = null;
	public WorldGraph wg = null;
	public SubZone mySubZone = null;
	public PoliceZone myPoliceZone = null;
	public int myZoneNumber;
	public int mySubZoneNumber;
	public ArrayList<Integer> lastIDs = new ArrayList<Integer>();
	public int tedadPoliceHa = -1;
	public int powNumForPolices = -1;
	public int refugePow = -1;
	// These Arrays could change to some trees for faster search
	public ArrayList<Building> reachableBuildings = new ArrayList<Building>();
	public ArrayList<Road> reachableRoads = new ArrayList<Road>();
	public ArrayList<Refuge> reachableRefuges = new ArrayList<Refuge>();
	// public TreeMap<Integer, Area> reachableAreas = new TreeMap<Integer,
	// Area>();
	// public TreeMap<Integer, Enterance> reachableEnterances = new
	// TreeMap<Integer, Enterance>();
	public List<Area> reachableAreas = new ArrayList<Area>();

	public ArrayList<Building> modelBuildings = new ArrayList<Building>();
	public ArrayList<Road> modelRoads = new ArrayList<Road>();
	public ArrayList<Refuge> modelRefuges = new ArrayList<Refuge>();
	public ArrayList<Area> modelAreas = new ArrayList<Area>();
	public ArrayList<FireBrigade> modelFireBrigades = new ArrayList<FireBrigade>();
	public ArrayList<PoliceForce> modelPoliceForces = new ArrayList<PoliceForce>();
	public ArrayList<AmbulanceTeam> modelAmbulanceTeams = new ArrayList<AmbulanceTeam>();
	public ArrayList<FireStation> modelFireStations = new ArrayList<FireStation>();
	public ArrayList<PoliceOffice> modelPoliceOffices = new ArrayList<PoliceOffice>();
	public ArrayList<AmbulanceCentre> modelAmbulanceCentres = new ArrayList<AmbulanceCentre>();
	protected Map<Integer, Integer> needHelpHeard = new HashMap<Integer, Integer>();
	public HashMap<Integer, Integer> allEntities = new HashMap<Integer, Integer>();
	public HashMap<Integer, Integer> allRanks = new HashMap<Integer, Integer>();;
	public int powNum = -1;
	int channel;
	public boolean noCommi = false;
	protected HashMap<Integer, Integer> nHP;
	public int counterForHelp = 0;
	public int counterForAmbHelp = 0;
	public int counterForUnHelp = 0;
	public int counterForAmbUnHelp = 0;
	public int counterForRescue = -1;
	public int counterForUnLoad = -1;
	public static int counterForBusiness = 0;
	public int counterForPoliceMap = 0;
	int centerForPolice;
	ArrayList<Integer> policeDecisionHeard1;
	ArrayList<Integer> ambDecisionHeard;
	protected HashMap<Integer, Integer> longnHP = new HashMap<Integer, Integer>();
	protected HashMap<Integer, Integer> uNHP = new HashMap<Integer, Integer>();
	protected HashMap<Integer, Integer> nAP = new HashMap<Integer, Integer>();
	protected HashMap<Integer, Integer> longnAP = new HashMap<Integer, Integer>();
	protected HashMap<Integer, Integer> uNAP = new HashMap<Integer, Integer>();
	protected HashMap<Integer, Integer> longnCP = new HashMap<Integer, Integer>();
	protected HashMap<Integer, Integer> longnBP = new HashMap<Integer, Integer>();
	private ArrayList<Integer> polices = new ArrayList<Integer>();
	ArrayList<AmbsCvs> groups;
	private Radar radar = null;
	EntityID lastRescuedObject = null;
	protected int lastRescuedTime = -1;
	protected int lastSaidUnLoadTime = -1;
	protected TreeSet<Integer> heardclearers = new TreeSet<Integer>();
	public Zone myZone1;
	public ArrayList<Integer> sentBlockades = new ArrayList<Integer>();
	public static final int MAX_DISTANCE_VIEW = 30000;
	public static final int MAX_DISTANCE_EXTINGUISH = 50000;
	public HashMap<Integer, Integer> policeFirstPositions = null;
	boolean wasItInLongnAP = false;
	boolean wasItInLongnHP = false;
	int areaPow = -1;
	int lastSaidPositionCycle = -1;
	public static final int cyclesBtweenSayingPos = 10;
	public static int MAX_GROUPING_DIST = 20000;
	protected SampleSearch search;
	public int mySubZoneID = -1;
	public boolean shouldBaseDecideForRandomWalk = true;
	public static final int LTSDORI = -50000000;
	public boolean movingToRefuge = false;

	public enum MapType {
		VC, Paris, Kobe, Berlin, Istanbul, Unknown
	}

	public MapType map = MapType.Unknown;

	protected void postConnect() {
		super.postConnect();
		System.out.println("Start of Precomputation");
		cycleTimer = new Timer();
		cycleTimer.setTime(0);
		if (psdConfig.get("Agent", "AgentLog", "LogToFile").equals("True"))
			logger = new Logger(psdConfig
					.get("Agent", "AgentLog", "LogAddress")
					+ "/"
					+ getName().replaceAll("agent.", "")
					+ "-"
					+ me().getID().getValue(), cycleTimer);
		else
			logger = new Logger(null, cycleTimer);

		for (StandardEntity next : model.getAllEntities()) {
			if (next instanceof Area)
				modelAreas.add((Area) next);
			if (next instanceof Refuge) {
				modelRefuges.add((Refuge) next);
				modelBuildings.add((Building) next);
			} else if (next instanceof Building)
				modelBuildings.add((Building) next);
			else if (next instanceof Road)
				modelRoads.add((Road) next);
			if (next instanceof FireBrigade)
				modelFireBrigades.add((FireBrigade) next);
			else if (next instanceof PoliceForce)
				modelPoliceForces.add((PoliceForce) next);
			else if (next instanceof AmbulanceTeam)
				modelAmbulanceTeams.add((AmbulanceTeam) next);
			else if (next instanceof FireStation)
				modelFireStations.add((FireStation) next);
			else if (next instanceof PoliceOffice)
				modelPoliceOffices.add((PoliceOffice) next);
			else if (next instanceof AmbulanceCentre)
				modelAmbulanceCentres.add((AmbulanceCentre) next);
		}

		System.out.println("Before sort");

		Collections.sort(modelPoliceForces, IDcomparator);
		Collections.sort(modelAreas, IDcomparator);
		Collections.sort(modelPoliceForces, IDcomparator);
		Collections.sort(modelPoliceOffices, IDcomparator);
		Collections.sort(modelFireBrigades, IDcomparator);
		Collections.sort(modelFireStations, IDcomparator);
		Collections.sort(modelAmbulanceTeams, IDcomparator);
		Collections.sort(modelAmbulanceCentres, IDcomparator);
		Collections.sort(modelRefuges, IDcomparator);

		refugePow = Radar.findPow(modelRefuges.size());

		int i = 0;

		for (Area area : modelAreas) {
			allEntities.put(area.getID().getValue(), i);
			allRanks.put(i, area.getID().getValue());
			i++;
		}

		areaPow = Radar.findPow(i - 1);

		for (PoliceForce policeForce : modelPoliceForces) {
			allEntities.put(policeForce.getID().getValue(), i);
			allRanks.put(i, policeForce.getID().getValue());
			i++;
		}
		for (PoliceOffice policeOffice : modelPoliceOffices) {
			allEntities.put(policeOffice.getID().getValue(), i);
			allRanks.put(i, policeOffice.getID().getValue());
			i++;
		}
		for (FireBrigade fireAgent : modelFireBrigades) {
			allEntities.put(fireAgent.getID().getValue(), i);
			allRanks.put(i, fireAgent.getID().getValue());
			i++;
		}
		for (FireStation fireStation : modelFireStations) {
			allEntities.put(fireStation.getID().getValue(), i);
			allRanks.put(i, fireStation.getID().getValue());
			i++;
		}
		for (AmbulanceTeam ambulanceTeam : modelAmbulanceTeams) {
			allEntities.put(ambulanceTeam.getID().getValue(), i);
			allRanks.put(i, ambulanceTeam.getID().getValue());
			i++;
		}
		for (AmbulanceCentre ambulanceCentre : modelAmbulanceCentres) {
			allEntities.put(ambulanceCentre.getID().getValue(), i);
			allRanks.put(i, ambulanceCentre.getID().getValue());
			i++;
		}

		System.out.println("After sort");

		tedadPoliceHa = modelPoliceForces.size();
		powNumForPolices = Radar.findPow(tedadPoliceHa);
		powNum = Radar.findPow(allEntities.size());

		System.out.println("Before Search Computation");
		search = new SampleSearch(model);
		neighbours = search.getGraph();
		System.out.println("After Search Computation");

		System.out.println("Start of WorldGraph Precomputation");
		wg = new WorldGraph(model, modelBuildings, modelRoads, modelRefuges);
		if (modelBuildings.size() == 735 && modelRoads.size() == 1515) { // MAP
			// BUGS:
			// KOBE
			removeConnectivity(31259, 31261);
			removeConnectivity(3930, 27209);
		}
		// VC1:
		if (modelBuildings.size() == 1258 && modelRoads.size() == 1954) {
			removeConnectivity(48405, 4264);
			removeConnectivity(49086, 53219);
			removeConnectivity(49086, 49087);
			removeConnectivity(48442, 7585);
		}
		log("BuildingsSize: " + modelBuildings.size());
		log("RoadsSize: " + modelRoads.size());
		log("OfflineFile: "
				+ psdConfig.get("Agent", "AgentLog", "OffLogFileName"));
		System.out.println("Precomputation finished");
		System.out.println("My ID: " + me().getID().getValue());
		if (1253 <= modelBuildings.size() && modelBuildings.size() <= 1273
				&& 1944 <= modelRoads.size() && modelRoads.size() <= 1964)
			map = MapType.VC;
		if (1608 <= modelBuildings.size() && modelBuildings.size() <= 1628
				&& 3015 <= modelRoads.size() && modelRoads.size() <= 3035)
			map = MapType.Paris;
		if (1416 <= modelBuildings.size() && modelBuildings.size() <= 1436
				&& 3375 <= modelRoads.size() && modelRoads.size() <= 3395)
			map = MapType.Berlin;
		if (726 <= modelBuildings.size() && modelBuildings.size() <= 746
				&& 1505 <= modelRoads.size() && modelRoads.size() <= 1525)
			map = MapType.Kobe;
		if (1234 <= modelBuildings.size() && modelBuildings.size() <= 1254
				&& 3327 <= modelRoads.size() && modelRoads.size() <= 3347)
			map = MapType.Istanbul;
		for (Area area : modelAreas)
			for (StandardEntity e : model.getObjectsInRange(area.getID(),
					MAX_DISTANCE_EXTINGUISH))
				if (e instanceof Area && !e.equals(area)) {
					double dist = Math.hypot(area.getX() - ((Area) e).getX(),
							area.getY() - ((Area) e).getY());

					if (map == MapType.Berlin || map == MapType.Paris
							|| map == MapType.Istanbul) {
						if (dist <= MAX_DISTANCE_EXTINGUISH) {
							if (e instanceof Road)
								area.nearRoads.add((Road) e);
							else if (e instanceof Building)
								area.nearBuildings.add((Building) e);
						}
					} else if (dist <= MAX_DISTANCE_VIEW) {
						if (e instanceof Road)
							area.nearRoads.add((Road) e);
						else if (e instanceof Building)
							area.nearBuildings.add((Building) e);
					}
					if (dist <= MAX_DISTANCE_EXTINGUISH
							&& area instanceof Building)
						area.nearAreas50000.add((Area) e);

				}
		setCenterOfMap();
		if (map.equals(MapType.Berlin) || map.equals(MapType.Paris)
				|| map.equals(MapType.Istanbul)) {
			STUCKCYCLES *= 2;
			positions = new int[STUCKCYCLES];
		}

		System.gc();
	}

	public void removeConnectivity(int id1, int id2) {
		worldGraph.Area area = ((Area) model.getEntityByInt(id1)).worldGraphArea;
		Enterance enterance = null;
		for (Enterance e : area.enterances)
			if (e.neighbourAreaInModel.getID().getValue() == id2) {
				enterance = e;
				break;
			}
		Enterance neighbour = enterance.neighbour;
		area.enterances.remove(enterance);
		neighbour.area.enterances.remove(neighbour);
	}

	int tedadByteChannel0 = 0;

	int tedad = 0;

	protected void selectChannel() {
		tedad = config.getIntValue("comms.channels.count");
		tedadByteChannel0 = config
				.getIntValue("comms.channels.0.messages.size");
		// System.out.println("tedad : " + tedad);
		if (tedad <= 1) {
			noCommi = true;
			channel = 0;
		}
		if (!noCommi) {
			channel = 1;
		}
		// TODO

		// float noise = -1;
		//
		// for(int i = 0 ; i < tedad ; i ++)
		// {
		// noise =
		// config.getIntValue("comms.channels.noise.1.output.dropout.p");
		// }
		//
		// System.out.println("noise : " + noise);
		// System.out.println("noCommi : " + noCommi);
		// System.out.println("channel : " + channel);
	}

	protected void selectCenterForPolice() {
		if (model.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE).size() > 0)
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE)) {
				centerForPolice = e.getID().getValue();
				break;
			}
		else if (model.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE)
				.size() > 0)
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE)) {
				centerForPolice = e.getID().getValue();
				break;
			}
		else if (model.getEntitiesOfType(StandardEntityURN.FIRE_STATION).size() > 0)
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.FIRE_STATION)) {
				centerForPolice = e.getID().getValue();
				break;
			}
		else if (model.getEntitiesOfType(StandardEntityURN.POLICE_FORCE).size() > 0) {
			ArrayList<Integer> policeForce = new ArrayList<Integer>();
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.POLICE_FORCE))
				policeForce.add(e.getID().getValue());
			Collections.sort(policeForce);
			centerForPolice = policeForce.get(0);
		} else if (model.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)
				.size() > 0) {
			ArrayList<Integer> ambTeam = new ArrayList<Integer>();
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM))
				ambTeam.add(e.getID().getValue());
			Collections.sort(ambTeam);
			centerForPolice = ambTeam.get(0);
		} else if (model.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)
				.size() > 0) {
			ArrayList<Integer> fireBrigade = new ArrayList<Integer>();
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE))
				fireBrigade.add(e.getID().getValue());
			Collections.sort(fireBrigade);
			centerForPolice = fireBrigade.get(0);
		} else
			centerForPolice = 0;
	}

	int centerForAmb;

	protected void selectCenterForAmb() {
		if (model.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE).size() > 0)
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE)) {
				centerForAmb = e.getID().getValue();
				break;
			}
		else if (model.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)
				.size() > 0) {
			ArrayList<Integer> ambTeam = new ArrayList<Integer>();
			for (Entity e : model
					.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM))
				ambTeam.add(e.getID().getValue());
			Collections.sort(ambTeam);
			centerForAmb = ambTeam.get(0);
		}
	}

	protected void sendPoliceMap(ArrayList<Integer> al) {
		BitSet bits = radar.mapHeader(Radar.HEADER_POLICEmAP);
		int start = Radar.headerSize;
		for (int i = 0; i < al.size(); i++)
			start = Radar.calcWithBitSet(bits, al.get(i) + 1, start,
					powNumForPolices);
		bits.set(start);
		if (Radar.codeToByteArray(bits).length > 1) {
			counterForPoliceMap++;
			byte[] byteArray = Radar.codeToByteArray(bits);
			sendSpeak(time, channel, byteArray);
			if (channel != 0)
				sendSpeak(time, 0, byteArray);
		}
	}

	ArrayList<Integer> lastHeardDecision = new ArrayList<Integer>();

	protected void hearPoliceMap() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_POLICEmAP) {
							policeDecisionHeard1 = radar.mapDecodeForPolice(
									recivedSpeak, powNumForPolices);
							haveIHeardFromRadar = true;
						}
					}
				}
			}
		}
		if (haveIHeardFromRadar
				&& !sameAL(lastHeardDecision, policeDecisionHeard1)) {
			if (needNewZoning(lastHeardDecision, policeDecisionHeard1)) {
				setMyZone(false);
				for (Zone z : zones)
					for (PoliceZone pz : z.policeZones)
						for (SubZone sz : pz.subZones) {
							for (Building b : sz.buildings)
								System.out.print(b.getID().getValue() + " ");
						}
				System.out.print("; ");
			}
			findMySubzone(policeDecisionHeard1);
		}
		lastHeardDecision = new ArrayList<Integer>();
		if (policeDecisionHeard1 != null)
			lastHeardDecision.addAll(policeDecisionHeard1);
	}

	protected void sendAmbMap(ArrayList<Integer> AL) {
		BitSet bits = new BitSet();
		bits = radar.mapHeader(Radar.HEADER_AMBmAP);
		int start = Radar.headerSize;
		for (int i = 0; i < AL.size() - 1; i += 2) {
			if (AL.get(i + 1) > 7)
				AL.set(i + 1, 7);
			int a = -1;
			if (AL.get(i) == -1)
				a = 0;
			else
				a = allEntities.get(AL.get(i)) + 1;
			start = radar.mapCodeForAmb(bits, a, AL.get(i + 1), start, areaPow);
		}
		bits.set(start);
		if (Radar.codeToByteArray(bits).length > 1) {
			byte[] byteArray = Radar.toByteArray(bits);
			sendSpeak(time, channel, byteArray);
			if (channel != 0)
				sendSpeak(time, 0, byteArray);
		}
	}

	protected void hearAmbMap() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_AMBmAP) {
							ambDecisionHeard = new ArrayList<Integer>();
							ArrayList<Integer> al = new ArrayList<Integer>();
							al.addAll(radar.mapDecodeForAmb(recivedSpeak,
									areaPow));
							for (int i = 0; i < al.size(); i += 2) {
								if (al.get(i) != -1) {
									ambDecisionHeard.add(allRanks
											.get(al.get(i)));
									ambDecisionHeard.add(al.get(i + 1));
								} else {
									ambDecisionHeard.add(al.get(i));
									ambDecisionHeard.add(al.get(i + 1));
								}
							}
						}
					}
				}
			}
		}
	}

	protected void sendCheckForOthersForNoCommi(int id) {
		BitSet bits = new BitSet();
		int start = 0;
		start = Radar.calcWithBitSet(bits, Radar.HEADER_CHECKfORaMB, start,
				Radar.headerSize);
		if (id == 0) {
			byte[] byteArray = Radar.toByteArray(bits);
			sendSpeak(time, channel, byteArray);
		} else {
			start = Radar.calcWithBitSet(bits, id, start, 20);
			byte[] byteArray = Radar.toByteArray(bits);
			sendSpeak(time, channel, byteArray);
		}
	}

	protected void sendCheckForOthersOfAmb(int id) {
		BitSet bits = new BitSet();
		int start = 0;
		start = Radar.calcWithBitSet(bits, Radar.HEADER_CHECKfORaMB, start,
				Radar.headerSize);
		if (id == 0) {
			byte[] byteArray = Radar.toByteArray(bits);
			if (counterForUnLoad < 1) {
				sendSpeak(time, channel, byteArray);
				lastSaidUnLoadTime = time;
			}
			sendSpeak(time, 0, byteArray);
		} else {
			start = Radar.calcWithBitSet(bits, id, start, 20);
			byte[] byteArray = Radar.toByteArray(bits);
			if (counterForRescue < 1)
				sendSpeak(time, channel, byteArray);
			sendSpeak(time, 0, byteArray);
		}
	}

	protected void hearCheckForOthersOfAmb() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						if (radar.tellHeader(recivedSpeak) == Radar.HEADER_CHECKfORaMB) {
							int id = cmd.getAgentID().getValue();
							AmbulanceTeam amb = (AmbulanceTeam) model
									.getEntityByInt(id);
							if (recivedSpeak.length == 1) {
								amb.target = -1;
								if (id == me().getID().getValue())
									counterForUnLoad++;
							} else {
								int hum = Radar.fromMabnaye2ToMabnaye10(Radar
										.fromByteArray(recivedSpeak), 24, 4);
								if (model.exist(hum)) {
									Human h = (Human) (model
											.getEntityByInt(hum));
									if (id > h.owner) {
										h.owner = id;
										amb.target = hum;
									}
									if (id == me().getID().getValue())
										counterForRescue++;
								} else {
									addCVlIDToModel(hum);
									Civilian cv = (Civilian) model
											.getEntityByInt(hum);
									cv.owner = amb.getID().getValue();
									amb.target = hum;
								}
							}
						}

					}
				}
			}
		}
	}

	protected void sendHelp() {
		Human me = (Human) me();
		int id = me.getID().getValue(), pos = me.getPosition().getValue();
		if (!longnHP.containsKey(id) || counterForHelp < 5) {
			lastSaidPositionCycle = time;
			byte[] byteArray = radar.code(Radar.HEADER_HELP, allEntities
					.get(pos));
			sendSpeak(time, channel, byteArray);
			if (channel != 0)
				sendSpeak(time, 0, byteArray);
		}
	}

	protected void unSendHelp() {
		Human me = (Human) me();
		int pos = me.getPosition().getValue();
		lastSaidPositionCycle = time;
		byte[] byteArray = radar
				.code(Radar.HEADER_UNhELP, allEntities.get(pos));
		sendSpeak(time, channel, byteArray);
		sendSpeak(time, 0, byteArray);
	}

	protected void sendHelpForAmb() {
		Human me = (Human) me();
		int id = me.getID().getValue();
		if (!longnAP.containsKey(id) || counterForAmbHelp < 5) {
			lastSaidPositionCycle = time;
			me.setDeadtime(time);
			int a = 0;
			if (me.isBuriednessDefined()) {
				if (me.getBuriedness() < 128)
					a = me.getBuriedness();
				else
					a = 127;
			} else
				a = 0;
			byte[] byteArray = radar.codeForAmbHelp(Radar.HEADER_AMBhELP,
					allEntities.get(me.getPosition().getValue()), me.deadtime,
					a);
			sendSpeak(time, channel, byteArray);
			if (channel != 0)
				sendSpeak(time, 0, byteArray);
		}
	}

	protected void unSendHelpForAmb() {
		Human me = (Human) me();
		int pos = me.getPosition().getValue();
		byte[] byteArray = radar.code(Radar.HEADER_AMBuNhELP, allEntities
				.get(pos));
		lastSaidPositionCycle = time;
		sendSpeak(time, channel, byteArray);
		sendSpeak(time, 0, byteArray);
	}

	protected Boolean checkCVInChangeSet(EntityID entityID) {
		for (EntityID eID : changeSet.getChangedEntities()) {
			if (eID == entityID) {
				return true;
			}
		}
		return false;
	}

	protected void sendCVlInformationForNoCommi() {
		if (noCommi == true) {
			int gonjayesh = (tedadByteChannel0 - 135) * 8 - 20 - powNum - 9 - 7
					- Radar.headerSize - 1;
			BitSet bits = new BitSet();
			int start = Radar.calcWithBitSet(bits, Radar.HEADER_CIVILIAN, 0,
					Radar.headerSize);
			for (Entity entit : model
					.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
				if (start < gonjayesh) {
					Civilian cv = (Civilian) entit;
					if (cv.isPositionDefined()) {
						start = Radar.calcWithBitSet(bits, cv.getID()
								.getValue(), start, 20);
						start = Radar.calcWithBitSet(bits, allEntities.get(cv
								.getPosition().getValue()), start, powNum);
						start = Radar.calcWithBitSet(bits, cv.deadtime, start,
								9);
						if (cv.isBuriednessDefined())
							start = Radar.calcWithBitSet(bits, cv
									.getBuriedness(), start, 7);
						else
							start = Radar.calcWithBitSet(bits, 0, start, 7);
					}
				} else
					break;
			}
			bits.set(start);
			byte[] byteArray = Radar.codeToByteArray(bits);
			if (byteArray.length > 1)
				sendSpeak(time, 0, byteArray);
		}
	}

	protected void sendCVlInformation() {
		Entity entity;
		ArrayList<Integer> properties = new ArrayList<Integer>();
		for (EntityID entityID : changeSet.getChangedEntities()) {
			entity = model.getEntity(entityID);
			if (entity instanceof Civilian) {
				Civilian cv = (Civilian) entity;
				if (((me() instanceof AmbulanceTeam
						&& lastRescuedObject != cv.getID() && (!longnCP
						.containsKey(cv.getID().getValue()) || longnCP.get(cv
						.getID().getValue()) != cv.getPosition().getValue())) && !modelAmbulanceTeams
						.contains(model.getEntityByInt(cv.getPosition()
								.getValue())))) {
					properties.add(entity.getID().getValue());
					properties.add(cv.getPosition().getValue());
					properties.add(cv.deadtime);
					cv.lastDeadTimeSaid = cv.deadtime;
					if (cv.isBuriednessDefined()) {
						if (cv.getBuriedness() < 128)
							properties.add(cv.getBuriedness());
						else
							properties.add(127);
					} else
						properties.add(0);
				}
				if (!(me() instanceof AmbulanceTeam)
						&& (!longnCP.containsKey(cv.getID().getValue()) || longnCP
								.get(cv.getID().getValue()) != cv.getPosition()
								.getValue())
						&& !modelAmbulanceTeams.contains(model
								.getEntityByInt(cv.getPosition().getValue()))) {
					if (cv.lastDeadTimeSaid < 0
							|| Math.abs(cv.lastDeadTimeSaid - cv.deadtime) > 100) {
						properties.add(entity.getID().getValue());
						properties.add(cv.getPosition().getValue());
						properties.add(cv.deadtime);
						cv.lastDeadTimeSaid = cv.deadtime;
						if (cv.isBuriednessDefined()) {
							if (cv.getBuriedness() < 128)
								properties.add(cv.getBuriedness());
							else
								properties.add(127);
						} else
							properties.add(0);
					}
				}
			}
		}
		if (properties.size() > 0) {
			BitSet bits = new BitSet();
			bits = radar.mapHeader(Radar.HEADER_CIVILIAN);
			int start = Radar.headerSize;
			int pos = 0;
			if (me() instanceof Human)
				pos = ((Human) me()).getPosition().getValue();
			else
				pos = me().getID().getValue();
			Radar.calcWithBitSet(bits, allEntities.get(pos), start, powNum);
			Area area = (Area) model.getEntityByInt(pos);
			int powNumNearBuildings = Radar
					.findPow(area.nearBuildings.size() + 1);
			start = start + powNum;
			for (int n = 0; n < properties.size(); n += 4) {
				Area a = (Area) model.getEntityByInt(properties.get(n + 1));
				if (area.nearBuildings.contains(a)) {
					start = radar.CVlCodeForID(bits, properties.get(n), start);
					start = radar
							.CVlCodeForPos(bits, area.nearBuildings
									.indexOf(model.getEntityByInt(properties
											.get(n + 1))) + 1, start,
									powNumNearBuildings);
					start = radar.CVlCodeForDeadTimeAndBN(bits, properties
							.get(n + 2), properties.get(n + 3), start);
				} else if (pos == a.getID().getValue() && a instanceof Building) {
					start = radar.CVlCodeForID(bits, properties.get(n), start);
					start = radar.CVlCodeForPos(bits, 0, start,
							powNumNearBuildings);
					start = radar.CVlCodeForDeadTimeAndBN(bits, properties
							.get(n + 2), properties.get(n + 3), start);
				}
			}
			bits.set(start);
			byte[] byteArray = Radar.codeToByteArray(bits);
			sendSpeak(time, channel, Radar.codeToByteArray(bits));
			lastSaidPositionCycle = time;
			if (channel != 0)
				sendSpeak(time, channel, byteArray);
		}
	}

	protected void sendBuildingInformation() {
		Entity entity;
		Building building;
		ArrayList<Integer> properties = new ArrayList<Integer>();
		for (EntityID entityID : changeSet.getChangedEntities()) {
			entity = model.getEntity(entityID);
			if (entity instanceof Building) {
				building = (Building) entity;
				if (building.isFierynessDefined()
						&& building.getFieryness() > 0)
					buildsOfFire.add(building);
			}
		}

		for (Building build : buildsOfFire) {
			int fireness = 0;
			int id = build.getID().getValue();
			if (build.isFierynessDefined()) {
				fireness = build.getFieryness();
			}
			if (!longnBP.containsKey(id) || longnBP.get(id) != fireness) {
				properties.add(id);
				properties.add(fireness);
			}
		}

		if (properties.size() > 0) {
			BitSet bits = new BitSet();
			ArrayList<Integer> buildingsThatAreNotInChangeSet = new ArrayList<Integer>();
			int start = 0;
			bits = radar.mapHeader(Radar.HEADER_BUILDING);
			start = Radar.headerSize;
			int pos = 0;
			if (me() instanceof Human)
				pos = ((Human) me()).getPosition().getValue();
			else
				pos = me().getID().getValue();
			start = Radar.calcWithBitSet(bits, allEntities.get(pos), start,
					radar.powNum);
			Area thisArea = (Area) model.getEntityByInt(pos);
			int powNumNearBuildings = Radar.findPow(thisArea.nearBuildings
					.size() + 1);
			for (int n = 0; n < properties.size(); n += 2) {
				if (thisArea.nearBuildings.contains(model
						.getEntityByInt(properties.get(n)))) {
					start = Radar.calcWithBitSet(bits,
							thisArea.nearBuildings.indexOf(model
									.getEntityByInt(properties.get(n))) + 1,
							start, powNumNearBuildings);
					start = Radar.calcWithBitSet(bits, properties.get(n + 1),
							start, Radar.tedadBitFireness);
					int lts = 0;
					if (((Building) model.getEntityByInt(properties.get(n))).lastTimeSeen > 0)
						lts = ((Building) model.getEntityByInt(properties
								.get(n))).lastTimeSeen;
					start = Radar.calcWithBitSet(bits, lts, start, 9);
				} else if (properties.get(n) == pos) {
					start = Radar.calcWithBitSet(bits, 0, start,
							powNumNearBuildings);
					start = Radar.calcWithBitSet(bits, properties.get(n + 1),
							start, Radar.tedadBitFireness);
					int lts = 0;
					if (((Building) model.getEntityByInt(properties.get(n))).lastTimeSeen > 0)
						lts = ((Building) model.getEntityByInt(properties
								.get(n))).lastTimeSeen;
					start = Radar.calcWithBitSet(bits, lts, start, 9);
				} else {
					buildingsThatAreNotInChangeSet.add(properties.get(n));
					buildingsThatAreNotInChangeSet.add(properties.get(n + 1));
				}

			}
			if (start > Radar.headerSize + powNum) {
				bits.set(start);
				byte[] byteArray = Radar.toByteArray(bits);
				lastSaidPositionCycle = time;
				sendSpeak(time, channel, byteArray);
				if (channel != 0)
					sendSpeak(time, 0, byteArray);
			}
			if (buildingsThatAreNotInChangeSet.size() > 0) {
				bits = new BitSet();
				bits = radar
						.mapHeader(Radar.HEADER_BuilDingOutOFreAchNeaRRoads);
				start = Radar.headerSize;
				for (int n = 0; n < buildingsThatAreNotInChangeSet.size(); n += 2) {
					int lts = 0;
					if (((Building) model
							.getEntityByInt(buildingsThatAreNotInChangeSet
									.get(n))).lastTimeSeen > 0)
						lts = ((Building) model
								.getEntityByInt(buildingsThatAreNotInChangeSet
										.get(n))).lastTimeSeen;
					start = radar.buildingCode(bits, allEntities
							.get(buildingsThatAreNotInChangeSet.get(n)),
							buildingsThatAreNotInChangeSet.get(n + 1), lts,
							start);
				}
				bits.set(start);
				byte[] byteArrayAD = Radar.toByteArray(bits);
				sendSpeak(time, channel, byteArrayAD);
				if (channel != 0)
					sendSpeak(time, 0, byteArrayAD);
			}
		}
	}

	protected int sendBlockades(Area _area, boolean[] isUsefull, BitSet bits,
			int start, int num, int areaPow) {
		worldGraph.Area area = _area.worldGraphArea;
		int a = area.enterances.size();
		int[][] array = new int[a][a];
		for (Enterance e : area.enterances)
			for (Enterance f : e.internalEnterances)
				array[e.id][f.id] = 1;
		for (Enterance e : area.enterances) {
			if (e.isItConnectedToNeighbour == true)
				array[e.id][e.id] = 1;
			else
				array[e.id][e.id] = 0;
		}
		int home = radar.calcForBlockade(array, isUsefull, bits, start, num,
				areaPow);
		return home;
	}

	protected void hearBlockades(byte[] byteArray, boolean isSelfHear, int IDon) {
		int ID = allRanks.get(radar.decodeIDForBlockade(byteArray));
		if (model.getEntityByInt(IDon) instanceof Human)
			addAgentPositionToModel(IDon, ID);
		Area _area = (Area) (model.getEntityByInt(ID));
		if (isSelfHear)
			return;
		BitSet bits = Radar.fromByteArray(byteArray);
		int lastBitPos = -1;
		for (lastBitPos = byteArray.length * 8 - 1; lastBitPos > 0; lastBitPos--)
			if (bits.get(lastBitPos) == true)
				break;
		bits.set(lastBitPos, false);
		int start = Radar.headerSize + powNum + 1;
		while (start < lastBitPos) {
			int thisID = Radar.fromMabnaye2ToMabnaye10(bits, start
					+ Radar.findPow(_area.nearRoads.size() + 1), start);
			int[][] array = radar.unCalcForBlockade(bits, _area, start, _area
					.getID().getValue(), logger, model);
			if (thisID == 0)
				thisID = _area.getID().getValue();
			else
				thisID = _area.nearRoads.get(thisID - 1).getID().getValue();
			Area thisArea = (Area) model.getEntityByInt(thisID);
			worldGraph.Area thisWorldGraphArea = thisArea.worldGraphArea;
			start = radar.shoro;
			if (isInChangeSet(thisArea.getID()))
				continue;
			for (int i = 0; i < array.length; i++)
				thisWorldGraphArea.enterances.get(i).internalEnterances.clear();
			for (int i = 0; i < array.length; i++)
				for (int j = i + 1; j < array.length; j++) {
					if (array[i][j] == 1) {
						thisWorldGraphArea.enterances.get(i).internalEnterances
								.add(thisWorldGraphArea.enterances.get(j));
						thisWorldGraphArea.enterances.get(j).internalEnterances
								.add(thisWorldGraphArea.enterances.get(i));
					}
				}
			for (int a = 0; a < array.length; a++) {
				boolean connectivity = (array[a][a] == 1) ? true : false;
				thisWorldGraphArea.enterances.get(a).isItConnectedToNeighbour = connectivity;
				thisWorldGraphArea.enterances.get(a).neighbour.isItConnectedToNeighbour = connectivity;
			}
			radar.sentBlockades.add(thisArea.getID().getValue());
		}
	}

	protected void hearBlockadesAD(byte[] byteArray, boolean isSelfHear) {
		if (isSelfHear)
			return;
		BitSet bits = Radar.fromByteArray(byteArray);
		int lastBitPos = -1;
		for (lastBitPos = byteArray.length * 8 - 1; lastBitPos > 0; lastBitPos--)
			if (bits.get(lastBitPos) == true)
				break;
		bits.set(lastBitPos, false);
		int start = Radar.headerSize + 1;
		while (start < lastBitPos) {
			int[][] array = radar.unCalcForBlockadeAD(bits, start, allRanks,
					model, logger);
			int thisID = allRanks.get(Radar.fromMabnaye2ToMabnaye10(bits, start
					+ powNum, start));
			Area thisArea = (Area) model.getEntityByInt(thisID);
			worldGraph.Area thisWorldGraphArea = thisArea.worldGraphArea;
			start = radar.shoroAD;
			if (isInChangeSet(thisArea.getID()))
				continue;
			for (int i = 0; i < array.length; i++)
				thisWorldGraphArea.enterances.get(i).internalEnterances.clear();
			for (int i = 0; i < array.length; i++)
				for (int j = i + 1; j < array.length; j++) {
					if (array[i][j] == 1) {
						thisWorldGraphArea.enterances.get(i).internalEnterances
								.add(thisWorldGraphArea.enterances.get(j));
						thisWorldGraphArea.enterances.get(j).internalEnterances
								.add(thisWorldGraphArea.enterances.get(i));
					}

				}
			for (int a = 0; a < array.length; a++) {
				boolean connectivity = (array[a][a] == 1) ? true : false;
				thisWorldGraphArea.enterances.get(a).isItConnectedToNeighbour = connectivity;
				thisWorldGraphArea.enterances.get(a).neighbour.isItConnectedToNeighbour = connectivity;
			}
			radar.sentBlockades.add(thisArea.getID().getValue());
		}
	}

	protected void sendNoWay(ArrayList<Integer> ids, Area myPosition, Boolean x) {
		BitSet bits = new BitSet();
		int start = 0;
		if (x == true) {
			lastSaidPositionCycle = time;
			start = Radar.calcWithBitSet(bits, Radar.HEADER_NOwAY, start,
					Radar.headerSize);
			start++;
			start = Radar.calcWithBitSet(bits, allEntities.get(myPosition
					.getID().getValue()), start, powNum);
			for (int i = 0; i < ids.size(); i++)
				start = Radar.calcWithBitSet(bits, ids.get(i), start, Radar
						.findPow(myPosition.nearRoads.size() + 1));

		} else {
			start = Radar.calcWithBitSet(bits, Radar.HEADER_NOwAY, start,
					Radar.headerSize);
			bits.set(start);
			start++;
			for (int i = 0; i < ids.size(); i++)
				start = Radar.calcWithBitSet(bits, allEntities.get(ids.get(i)),
						start, powNum);
		}
		bits.set(start);
		byte[] byteArray = Radar.codeToByteArray(bits);
		sendSpeak(time, channel, byteArray);
		if (channel != 0)
			sendSpeak(time, 0, byteArray);
	}

	protected void hearNoWay() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					ArrayList<Integer> IDs = new ArrayList<Integer>();
					if (recivedSpeak.length > 0) {
						if (radar.tellHeader(recivedSpeak) == Radar.HEADER_NOwAY) {
							if (radar.tellHeaderForPC(recivedSpeak) == true) {
								IDs = radar.decodeIDForNoWay(recivedSpeak,
										model, allRanks);
								addAgentPositionToModel(cmd.getAgentID()
										.getValue(), IDs.get(0));
								;
								for (int i = 1; i < IDs.size(); i++) {
									int ID = IDs.get(i);
									if (!changeSet.getChangedEntities()
											.contains(
													model.getEntityByInt(ID)
															.getID())) {
										Area _area = (Area) model
												.getEntityByInt(ID);
										worldGraph.Area area = _area.worldGraphArea;
										for (Enterance enterance : area.enterances) {
											enterance.internalEnterances
													.clear();
											enterance.isItConnectedToNeighbour = enterance.neighbour.isItConnectedToNeighbour = false;
										}
										radar.sentBlockades.add(ID);
									}
								}
							} else if (radar.tellHeaderForPC(recivedSpeak) == false) {
								IDs = radar.decodeIDForNoWayAD(recivedSpeak,
										model, allRanks);
								for (int i = 0; i < IDs.size(); i++) {
									int ID = IDs.get(i);
									if (!changeSet.getChangedEntities()
											.contains(
													model.getEntityByInt(ID)
															.getID())) {
										Area _area = (Area) model
												.getEntityByInt(ID);
										worldGraph.Area area = _area.worldGraphArea;
										for (Enterance enterance : area.enterances) {
											enterance.internalEnterances
													.clear();
											enterance.isItConnectedToNeighbour = enterance.neighbour.isItConnectedToNeighbour = false;
										}
										radar.sentBlockades.add(ID);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	protected void checkWays() {
		if (me() instanceof Human) {
			ArrayList<Integer> sendNoWayIDs = new ArrayList<Integer>();
			ArrayList<Integer> sendNoWayAD = new ArrayList<Integer>();
			Human me = (Human) me();
			int start = 0;
			BitSet bits = new BitSet();
			ArrayList<Entity> notInNearestButInChangeSet = new ArrayList<Entity>();
			int pos = me.getPosition().getValue();
			start = Radar.calcWithBitSet(bits, Radar.HEADER_BLOCKADE, start,
					Radar.headerSize);
			start = Radar.calcWithBitSet(bits, 0, start, 1);
			start = Radar.calcWithBitSet(bits, allEntities.get(me.getPosition()
					.getValue()), start, powNum);
			Area _area = (Area) (model.getEntityByInt(pos));
			int areaPow = Radar.findPow(_area.nearRoads.size() + 1);
			for (EntityID entityID : changeSet.getChangedEntities()) {
				if (model.getEntityByInt(entityID.getValue()) instanceof Road) {
					Entity entit = model.getEntityByInt(entityID.getValue());
					if (_area.nearRoads.contains(entit)
							&& !sentBlockades
									.contains(entit.getID().getValue())) {
						int num = _area.nearRoads.indexOf(entit);
						boolean[] isUseful = new boolean[1];
						isUseful[0] = true;
						start = sendBlockades((Area) entit, isUseful, bits,
								start, num + 1, areaPow);
						if (isUseful[0] == false)
							sendNoWayIDs.add(num + 1);
					} else if (!sentBlockades
							.contains(entit.getID().getValue())) {
						if (entit.getID().getValue() != me.getPosition()
								.getValue())
							notInNearestButInChangeSet.add(entit);
						else {
							boolean[] isUseful = new boolean[1];
							isUseful[0] = true;
							start = sendBlockades((Area) entit, isUseful, bits,
									start, 0, areaPow);
							if (isUseful[0] == false)
								sendNoWayIDs.add(0);
						}
					}
				}
			}
			bits.set(start);
			if (start > Radar.headerSize + powNum) {
				byte[] byteArray = Radar.toByteArray(bits);
				sendSpeak(time, channel, byteArray);
				lastSaidPositionCycle = time;
				if (channel != 0)
					sendSpeak(time, 0, byteArray);
			}
			if (notInNearestButInChangeSet.size() > 0) {
				bits = new BitSet();
				start = Radar.calcWithBitSet(bits, Radar.HEADER_BLOCKADE, 0,
						Radar.headerSize);
				start = Radar.calcWithBitSet(bits, 1, start, 1);
				boolean[] isUsefull = new boolean[1];
				for (Entity entit : notInNearestButInChangeSet) {
					isUsefull[0] = true;
					worldGraph.Area area_ = ((Area) entit).worldGraphArea;
					int a = area_.enterances.size();
					int[][] array = new int[a][a];
					for (Enterance e : area_.enterances)
						for (Enterance f : e.internalEnterances)
							array[e.id][f.id] = 1;
					for (Enterance e : area_.enterances) {
						if (e.isItConnectedToNeighbour == true)
							array[e.id][e.id] = 1;
						else
							array[e.id][e.id] = 0;
					}
					Radar.mark(array);
					int home = 0;
					BitSet jaSet = new BitSet();
					for (int b = 0; b < array.length; b++) {
						for (int i = b + 1; i < array.length; i++) {
							if (array[b][i] != 2) {
								if (array[b][i] == 1)
									jaSet.set(home);
								home++;
							}
						}
					}

					for (int b = 0; b < array.length; b++) {
						if (array[b][b] == 1)
							jaSet.set(home);
						home++;
					}

					int tedad1Ha = 0;
					for (int b = 0; b < home; b++)
						if (jaSet.get(b) == true)
							tedad1Ha++;

					if (tedad1Ha == home) {
						isUsefull[0] = false;
					} else {
						if (tedad1Ha == 0) {
							sendNoWayAD.add(entit.getID().getValue());
						} else {
							start = Radar.calcWithBitSet(bits, allEntities
									.get(entit.getID().getValue()), start,
									powNum);
							for (int i = 0; i < home; i++)
								if (jaSet.get(i) == true)
									bits.set(start + i);
							start += home;
						}
					}
				}
				bits.set(start);
				if (start > Radar.headerSize) {
					byte[] byteArray = Radar.toByteArray(bits);
					sendSpeak(time, channel, byteArray);
					if (channel != 0)
						sendSpeak(time, 0, byteArray);
				}
			}
			if (sendNoWayIDs.size() > 0)
				sendNoWay(sendNoWayIDs, _area, true);
			if (sendNoWayAD.size() > 0) {
				sendNoWay(sendNoWayAD, _area, false);
			}
		}
	}

	protected void hearWay() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						if (radar.tellHeader(recivedSpeak) == Radar.HEADER_BLOCKADE)
							if (radar.tellHeaderForPC(recivedSpeak) == true)
								hearBlockades(recivedSpeak, cmd.getAgentID()
										.equals(me().getID()), cmd.getAgentID()
										.getValue());
							else if (radar.tellHeaderForPC(recivedSpeak) == false)
								hearBlockadesAD(recivedSpeak, cmd.getAgentID()
										.equals(me().getID()));
					}
				}
			}
		}
	}

	protected void hearHelp() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_HELP) {
							nHP = new HashMap<Integer, Integer>();
							int id = cmd.getAgentID().getValue(), pos = allRanks
									.get(radar.decode(recivedSpeak));
							nHP.put(id, pos);
							longnHP.put(id, pos);
							addAgentPositionToModel(id, pos);
							if (id == me().getID().getValue())
								counterForHelp++;
						}
					}
				}
			}
		}
	}

	protected void unHearHelp() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_UNhELP) {
							uNHP = new HashMap<Integer, Integer>();
							int id = cmd.getAgentID().getValue(), pos = allRanks
									.get(radar.decode(recivedSpeak));
							uNHP.put(id, pos);
							addAgentPositionToModel(id, pos);
							if (id == me().getID().getValue())
								counterForUnHelp++;
						}
					}
				}
			}
		}
		for (Entry<Integer, Integer> entry : uNHP.entrySet()) {
			if (longnHP.containsKey(entry.getKey()))
				longnHP.remove(entry.getKey());
		}
	}

	protected void hearHelpForAmb() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_AMBhELP) {
							nAP = new HashMap<Integer, Integer>();
							int id = entity.getID().getValue();
							ArrayList<Integer> IDP = radar
									.decodeForAmbHelp(recivedSpeak);
							Human human = ((Human) entity);
							int pos = allRanks.get(IDP.get(0));
							human.deadtime = IDP.get(1);
							human.setBuriedness(IDP.get(2));
							nAP.put(id, pos);
							longnAP.put(id, pos);
							((Human) entity).hasBuriedness = true;
							if (id == me().getID().getValue())
								counterForAmbHelp++;
							addAgentPositionToModel(id, pos);
						}
					}
				}
			}
		}
	}

	protected void unHearHelpForAmb() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_AMBuNhELP) {
							uNAP = new HashMap<Integer, Integer>();
							int id = cmd.getAgentID().getValue(), pos = allRanks
									.get(radar.decode(recivedSpeak));
							uNAP.put(id, pos);
							addAgentPositionToModel(id, pos);
							if (id == me().getID().getValue())
								counterForAmbUnHelp++;
						}
					}
				}
			}
		}
		for (Entry<Integer, Integer> entry : uNAP.entrySet()) {
			int id = entry.getKey();
			if (longnHP.containsKey(id))
				longnAP.remove(id);
			((Human) model.getEntityByInt(id)).hasBuriedness = false;
		}
	}

	protected void hearCVlInformation() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_CIVILIAN) {
							ArrayList<Integer> al = new ArrayList<Integer>();
							al.addAll(radar.CVlDecode(recivedSpeak, model,
									allRanks, noCommi));
							if (model.getEntityByInt(cmd.getAgentID()
									.getValue()) instanceof Human
									&& !noCommi)
								addAgentPositionToModel(cmd.getAgentID()
										.getValue(), al.get(0));
							for (int i = 1; i < al.size(); i += 4) {
								int id = al.get(i);
								int pos = al.get(i + 1);
								if ((model.getEntityByInt(id) == null || !changeSet
										.getChangedEntities().contains(
												model.getEntityByInt(id)
														.getID()))
										&& pos != me().getID().getValue()) {
									int deadTime = al.get(i + 2);
									int BN = al.get(i + 3);
									if (cmd.getAgentID().equals(me().getID()) == false) {
										Entity e = model.getEntityByInt(id);
										if (e == null
												|| (e != null && checkCVInChangeSet(e
														.getID()) == false)) {
											addCVlToModel(id, pos, deadTime, BN);
										}
									}
								}
								longnCP.put(id, pos);
							}
						}
					}
				}
			}
		}
	}

	protected void addCVlIDToModel(int num) {
		Entity existingEntity = model.getEntityByInt(num);
		if (existingEntity == null) {
			EntityID e = new EntityID(num);
			existingEntity = Registry.getCurrentRegistry().createEntity(
					StandardEntityURN.CIVILIAN.toString(), e);
			model.addEntity(existingEntity);
		}
	}

	protected void addCVlToModel(int num, int pos, int deadTime, int BN) {
		Entity existingEntity = model.getEntityByInt(num);
		if (existingEntity == null) {
			EntityID e = new EntityID(num);
			existingEntity = Registry.getCurrentRegistry().createEntity(
					StandardEntityURN.CIVILIAN.toString(), e);
			model.addEntity(existingEntity);
		}
		if (existingEntity != null
				&& !(changeSet.getChangedEntities().contains(existingEntity))) {
			Entity e = model.getEntityByInt(pos);
			((Civilian) existingEntity).setPosition(e.getID());
			((Civilian) existingEntity).deadtime = deadTime;
			((Civilian) existingEntity).setBuriedness(BN);
			if (e instanceof Area) {
				((Civilian) existingEntity).setX(((Area) e).getX());
				((Civilian) existingEntity).setY(((Area) e).getY());
			}
		}
	}

	protected void addAgentPositionToModel(int id, int pos) {
		Entity existingEntity = model.getEntityByInt(id);
		if (existingEntity != null
				&& !(changeSet.getChangedEntities().contains(existingEntity))
				&& existingEntity != me())
			((Human) existingEntity).setPosition(model.getEntityByInt(pos)
					.getID());
	}

	protected void addBuildingToModel(int building, int firyness) {
		Entity existingEntity = model.getEntityByInt(building);
		if (existingEntity != null
				&& !(changeSet.getChangedEntities().contains(existingEntity))) {
			((Building) existingEntity).setFieryness(firyness);
			if (me() instanceof FireBrigade
					&& ((Building) existingEntity).isOnFire())
				for (Building b : ((Building) existingEntity).nearBuildings)
					if (b.lastTimeSeen < 0
							&& (!b.isFierynessDefined() || b.getFieryness() == 0)) {
						b.lastTimeSeen = LTSDORI;
					}
		}
	}

	protected void hearBuildingInformation() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_BUILDING) {
							int id, fireness, lts;
							ArrayList<Integer> al = new ArrayList<Integer>();
							al.addAll(radar.BuildingDecode(recivedSpeak, model,
									allRanks));
							if (model.getEntityByInt(cmd.getAgentID()
									.getValue()) instanceof Human)
								addAgentPositionToModel(cmd.getAgentID()
										.getValue(), al.get(0));
							for (int i = 1; i < al.size(); i += 3) {
								id = al.get(i);
								fireness = al.get(i + 1);
								lts = al.get(i + 2);
								if (cmd.getAgentID().equals(me().getID()) == false) {
									Building e = (Building) model
											.getEntityByInt(id);
									if (isInChangeSet(e.getID()) == false
											&& lts > e.lastTimeSeen) {
										addBuildingToModel(id, fireness);
										((Building) model.getEntityByInt(id)).lastTimeSeen = lts;
									}
								}
								longnBP.put(id, fireness);
							}
						} else if (header == Radar.HEADER_BuilDingOutOFreAchNeaRRoads) {
							int id, fireness, lts;
							ArrayList<Integer> al = new ArrayList<Integer>();
							al.addAll(radar
									.BuildingDecodeOutOfReach(recivedSpeak));
							for (int i = 1; i < al.size(); i += 3) {
								id = allRanks.get(al.get(i));
								fireness = al.get(i + 1);
								lts = al.get(i + 2);
								if (cmd.getAgentID().equals(me().getID()) == false) {
									Building e = (Building) model
											.getEntityByInt(id);
									if (isInChangeSet(e.getID()) == false
											&& lts > e.lastTimeSeen) {
										addBuildingToModel(id, fireness);
										((Building) model.getEntityByInt(id)).lastTimeSeen = lts;
									}
								}
								longnBP.put(id, fireness);
							}
						}
					}
				}
			}
		}
	}

	protected void sendPolicePC(int zoneNum, int howImportant) {
		BitSet bits = new BitSet();
		int start = 0;
		start = Radar.calcWithBitSet(bits, Radar.HEADER_PolicePC, start,
				Radar.headerSize);
		start = Radar.calcWithBitSet(bits, zoneNum, start, powNumForPolices);
		start = Radar.calcWithBitSet(bits, howImportant, start, 7);
		byte[] byteArray = Radar.codeToByteArray(bits);
		sendSpeak(time, channel, byteArray);
		if (channel != 0)
			sendSpeak(time, 0, byteArray);
	}

	ArrayList<PolicePC> allPpcs = new ArrayList<PolicePC>();

	protected void hearPolicePC() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_PolicePC) {
							BitSet bits = Radar.fromByteArray(recivedSpeak);
							int headerSize = Radar.headerSize;
							int id = cmd.getAgentID().getValue();
							int zoneNum = Radar.fromMabnaye2ToMabnaye10(bits,
									headerSize + powNumForPolices, headerSize);
							int howImportant = Radar.fromMabnaye2ToMabnaye10(
									bits, headerSize + powNumForPolices + 7,
									headerSize + powNumForPolices);
							PolicePC ppc = new PolicePC();
							ppc.id = id;
							ppc.zoneNum = zoneNum;
							ppc.howImportant = howImportant;
							allPpcs.add(ppc);
						}
					}
				}
			}
		}
	}

	protected void setPolices() {
		for (PoliceForce p : modelPoliceForces) {
			if (!p.isHPDefined() || (p.isHPDefined() && p.getHP() != 0)
					&& !p.hasBuriedness) {
				boolean shouldAdd = true;
				for (int f : longnAP.keySet())
					if (p.getID().getValue() == f) {
						shouldAdd = false;
						break;
					}
				if (shouldAdd)
					polices.add(p.getID().getValue());
			}
		}
	}

	ArrayList<Integer> subzoneIDs = new ArrayList<Integer>();

	int lastTimePFmapSend = -1;

	protected void devotingSubzones() {
		subzoneIDs = new ArrayList<Integer>();
		if (policeFirstPositions == null) {
			policeFirstPositions = new HashMap<Integer, Integer>();
			for (PoliceForce pf : modelPoliceForces)
				policeFirstPositions.put(pf.getID().getValue(), pf
						.getPosition().getValue());
		}
		IDs = new ArrayList<Integer>();
		int numberOfZones = 5;
		for (PoliceForce policeForce : modelPoliceForces)
			if (!policeForce.hasBuriedness
					&& (!policeForce.isHPDefined() || (policeForce
							.isHPDefined() && policeForce.getHP() > 50)))
				IDs.add(policeForce.getID().getValue());
		numberOfZones = Math.min(5, (IDs.size() + 1) / 2);
		System.out.println("Sizes: " + lastIDs.size() + " " + IDs.size());
		if (lastIDs.size() == IDs.size() && sameAL(lastIDs, IDs)) {
			if (counterForPoliceMap > 3 && (time - lastTimePFmapSend) < 10) {
				return;
			}
		} else if (counterForPoliceMap > 3)
			counterForPoliceMap = 0;

		lastIDs = IDs;
		// zoning the map
		zoningForCenter(numberOfZones);

		// subZoning my PoliceZone...
		for (Zone z : zones)
			policeZoning(z.zoneNumber, z.policeZones.length);

		ArrayList<Integer> freePolices = new ArrayList<Integer>();
		freePolices.addAll(IDs);

		done: while (true)
			for (Zone z : zones)
				for (PoliceZone pz : z.policeZones) {
					if (freePolices.size() == 0)
						break done;
					int nearest = nearestPolice(pz.center.getX(), pz.center
							.getY(), freePolices, false);
					pz.policesDedicatedto.add(nearest);
					freePolices.remove(freePolices.indexOf(nearest));
				}
		for (Zone z : zones)
			for (PoliceZone pz : z.policeZones) {
				pz.setSubZonesForPolice(pz.policesDedicatedto.size());
				subZoningForPolice(z.zoneNumber, pz.policeZoneNumber,
						pz.subZones.length);
			}

		setIDsOfSubZones();
		setPriorityOfSubZones();
		Collections.sort(subzones, new Comparator<SubZone>() {
			public int compare(SubZone arg0, SubZone arg1) {
				if (arg0.priority > arg1.priority)
					return -1;
				if (arg0.priority < arg1.priority)
					return 1;
				return 0;
			}
		});
		freePolices = new ArrayList<Integer>();
		freePolices.addAll(IDs);
		int a[] = new int[modelPoliceForces.size()];
		for (int i = 0; i < modelPoliceForces.size(); i++)
			a[i] = -1;

		for (SubZone sz : subzones) {
			int nearestPF = nearestPolice(sz.center.getX(), sz.center.getY(),
					freePolices, false);
			int policeIndex = modelPoliceForces.indexOf(model
					.getEntityByInt(nearestPF));
			a[policeIndex] = sz.subZoneID;
			freePolices.remove(freePolices.indexOf(nearestPF));
		}
		for (int szID : a)
			subzoneIDs.add(szID);
		lastTimePFmapSend = time;
		sendPoliceMap(subzoneIDs);
	}

	private void setIDsOfSubZones() {
		subzones = new ArrayList<SubZone>();
		int counter = -1;
		for (Zone z : zones) {
			for (PoliceZone pz : z.policeZones) {
				for (SubZone sz : pz.subZones) {
					counter++;
					sz.setSubZoneID(counter);
					subzones.add(sz);
				}
			}
		}

	}

	private void setPriorityOfSubZones() {
		HashSet<Integer> nhpPositions = new HashSet<Integer>();
		for (int i : longnHP.values())
			nhpPositions.add(i);
		for (int NHPpos : nhpPositions) {
			Entity nhpEntity = model.getEntityByInt(NHPpos);
			float theta1 = (new Line(centerOfMap, new Point(((Area) nhpEntity)
					.getX(), ((Area) nhpEntity).getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta1 < 0)
				theta1 += 360;
			numberOfZones = Math.min(5, (IDs.size() + 1) / 2);
			int zoneNumber = (int) (theta1 / (360.0 / numberOfZones));

			float theta2 = (new Line(zones[zoneNumber].center, new Point(
					((Area) nhpEntity).getX(), ((Area) nhpEntity).getY())))
					.getTheta()
					* Degree.RAD2DEG;
			if (theta2 < 0)
				theta2 += 360;
			int policeZoneNumber = (int) (theta2 / (360.0 / zones[zoneNumber].policeZones.length));

			float theta3 = (new Line(
					zones[zoneNumber].policeZones[policeZoneNumber].center,
					new Point(((Area) nhpEntity).getX(), ((Area) nhpEntity)
							.getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta3 < 0)
				theta3 += 360;
			int subZoneNumber = (int) (theta3 / (360.0 / zones[zoneNumber].policeZones[policeZoneNumber].subZones.length));
			for (Entry<Integer, Integer> thisNHP : longnHP.entrySet()) {
				if (thisNHP.getValue() == NHPpos)
					if ((model.getEntityByInt(thisNHP.getKey())) instanceof FireBrigade)
						zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber].priority += 2;
					else
						zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber].priority++;
			}
		}

	}

	private void zoningForCenter(int zonesCount) {
		zones = new Zone[zonesCount];
		// pCount = tedade hadeghal agent ha baraye 1 zone
		int agentsCount = IDs.size();
		int pCount = agentsCount / zonesCount;
		// setting subzones...
		// setting policeZones...
		pCount = Math.min(2, pCount);
		for (int i = 0; i < zones.length; i++) {
			zones[i] = new Zone(i);
			if (pCount == 2) {
				zones[i].setPoliceZones(2);
				zones[i].policeZones[0] = new PoliceZone(0);
				zones[i].policeZones[1] = new PoliceZone(1);
			} else if (i < agentsCount % zonesCount) {
				zones[i].setPoliceZones(pCount + 1);
				for (int j = 0; j <= pCount; j++)
					zones[i].policeZones[j] = new PoliceZone(j);
			} else {
				zones[i].setPoliceZones(pCount);
				for (int v = 0; v < pCount; v++)
					zones[i].policeZones[v] = new PoliceZone(v);
			}
		}
		// dividing the whole map into Zones
		for (Area area : modelAreas) {
			float theta = (new Line(centerOfMap, new Point(area.getX(), area
					.getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta < 0)
				theta += 360;

			int zoneNumber = (int) (theta / (360.0 / zonesCount));
			area.worldGraphArea.ZoneNumberForPolice = zoneNumber;
			zones[zoneNumber].areas.add(area);
			if (area instanceof Refuge)
				zones[zoneNumber].refuges.add((Refuge) area);
			else if (area instanceof Building)
				zones[zoneNumber].buildings.add((Building) area);
			else if (area instanceof Road)
				zones[zoneNumber].roads.add((Road) area);
		}
		/**
		 * for (Zone z : zones) { int sumX = 0; int sumY = 0; for (Area area :
		 * z.areas) { sumX += area.getX(); sumY += area.getY(); } z.center = new
		 * Point(sumX / z.areas.size(), sumY / z.areas.size()); }
		 */
		for (Zone z : zones) {
			int minX = Integer.MAX_VALUE, maxX = Integer.MIN_VALUE;
			int minY = Integer.MAX_VALUE, maxY = Integer.MIN_VALUE;
			for (Area area : z.areas) {
				if (area.getX() > maxX)
					maxX = area.getX();
				if (area.getX() < minX)
					minX = area.getX();
				if (area.getY() > maxY)
					maxY = area.getY();
				if (area.getY() < minY)
					minY = area.getY();
			}
			z.center = new Point((minX + maxX) / 2, (minY + maxY) / 2);
		}
	}

	protected void centerForPolice() {
		if (time > 3)
			devotingSubzones();
	}

	ArrayList<Human> Agents = null;

	// AmbulanceTeam
	public void groupingForAmb() {
		Agents = new ArrayList<Human>();
		for (AmbulanceTeam amb : modelAmbulanceTeams) {
			// if (amb.target == -1)
			amb.ambGroup = false;
			amb.freetime = 0;
			amb.curPos = amb.getPosition();
			// if (amb.target == -1) {
			// amb.firstFreeTime = 0;
			// }// else {
			/*
			 * Human h = (Human) model.getEntityByInt(amb.target); if (h
			 * instanceof Civilian) { Civilian c = (Civilian) h; if
			 * (c.isBuriednessDefined() && c.getBuriedness() > 0) {
			 * amb.firstFreeTime = c.getBuriedness() + 20; if
			 * (c.isPositionDefined()) amb.curPos = c.getPosition(); } else {
			 * amb.firstFreeTime = 20; amb.curPos = amb.getPosition(); } } else
			 * if (h.isBuriednessDefined()) { amb.curPos = h.getPosition();
			 * amb.firstFreeTime = h.getBuriedness(); } }
			 */
		}

		for (worldGraph.Area area : wg.areas) {
			area.mark = false;
			area.agents = new ArrayList<Human>();
			area.civilians = new ArrayList<Civilian>();
			area.ambulances = new ArrayList<AmbulanceTeam>();
		}
		groups = new ArrayList<AmbsCvs>();

		for (AmbulanceTeam amb : modelAmbulanceTeams) {
			if (((amb.isHPDefined() && amb.getHP() > 0) || !amb.isHPDefined())) {
				worldGraph.Area ambArea = ((Area) model.getEntity(amb
						.getPosition())).worldGraphArea;
				if (amb.hasBuriedness == false && amb.ambGroup == false
						&& amb.target == -1) {
					ambArea.ambulances.add(amb);
				} else if (amb.hasBuriedness == true) {
					Area a = (Area) model.getEntity(amb.getPosition());
					if (a instanceof Building) {
						Building b = (Building) a;
						if (b.isOnFire() == false) {
							Agents.add(amb);
							ambArea.agents.add(amb);
						}
					}
				}
			}
		}

		for (PoliceForce p : modelPoliceForces)
			if (p.owner == -1
					&& ((p.isHPDefined() && p.getHP() > 0) || !p.isHPDefined())
					&& p.isBuriednessDefined() && p.getBuriedness() > 0) {
				Area a = (Area) model.getEntity(p.getPosition());
				if (a instanceof Building) {
					Building b = (Building) a;
					if (b.isOnFire() == false) {
						Agents.add(p);
						worldGraph.Area policeArea = ((Area) model.getEntity(p
								.getPosition())).worldGraphArea;
						policeArea.agents.add(p);
					}
				}
			}

		for (FireBrigade f : modelFireBrigades)
			if (f.owner == -1
					&& ((f.isHPDefined() && f.getHP() > 0) || !f.isHPDefined())
					&& f.isBuriednessDefined() && f.getBuriedness() > 0) {
				Area a = (Area) model.getEntity(f.getPosition());
				if (a instanceof Building) {
					Building b = (Building) a;
					if (b.isOnFire() == false) {
						Agents.add(f);
						worldGraph.Area fireArea = ((Area) model.getEntity(f
								.getPosition())).worldGraphArea;
						fireArea.agents.add(f);
					}
				}
			}

		for (StandardEntity entity : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian CVlian = (Civilian) entity;
			if (CVlian.isPositionDefined()) {
				Entity e = model.getEntity(CVlian.getPosition());
				if (CVlian.owner == -1
						&& ((CVlian.isHPDefined() && CVlian.getHP() > 0) || !CVlian
								.isHPDefined()) && e instanceof Building
						&& CVlian.deadtime > time + CVlian.getBuriedness() + 20
						&& CVlian.isBuriednessDefined()
						&& CVlian.getBuriedness() > 0) {
					Building b = (Building) e;
					if (b.isOnFire() == false) {
						if (e instanceof Area && !(e instanceof Refuge)) {
							worldGraph.Area cvarea = ((Area) e).worldGraphArea;
							cvarea.civilians.add(CVlian);
							log("Civilian: " + CVlian.getID().getValue());
						}
					}
				}
			}
		}

		for (AmbulanceTeam amb : modelAmbulanceTeams) {
			if (amb.ambGroup == false && amb.hasBuriedness == false)
				log("AMB: " + amb.getID().getValue());
			if (/*
				 * amb.target == -1 &&
				 */amb.ambGroup == false
					&& amb.hasBuriedness == false
					&& ((amb.isHPDefined() && amb.getHP() > 0) || !amb
							.isHPDefined())) {
				worldGraph.Area area = ((Area) model.getEntityByInt(amb
						.getPosition().getValue())).worldGraphArea;
				groups.add(getMinPathForAmb(area.enterances));
			}
		}
	}

	public AmbsCvs getMinPathForAmb(ArrayList<Enterance> start) {
		AmbsCvs a = new AmbsCvs();
		ArrayList<Enterance> layer = new ArrayList<Enterance>();

		wg.clearAreas();

		for (Enterance enterance : start) {
			if (!enterance.mark && !enterance.area.mark) {
				if (!(enterance.area.modelArea instanceof Refuge))
					a.cvs.addAll(enterance.area.civilians);
				a.ambs.addAll(enterance.area.ambulances);
				a.agents.addAll(enterance.area.agents);
				if (enterance.area.modelArea instanceof Refuge)
					a.refuges.add((Refuge) enterance.area.modelArea);
				for (AmbulanceTeam amb : enterance.area.ambulances)
					amb.ambGroup = true;
				enterance.area.mark = true;
				enterance.mark = true;
				layer.add(enterance);
			} else if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}
		}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer) {
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark
						&& !enterance.neighbour.area.mark) {
					Enterance neighbour = enterance.neighbour;
					if (!(neighbour.area.modelArea instanceof Refuge))
						a.cvs.addAll(neighbour.area.civilians);
					a.ambs.addAll(neighbour.area.ambulances);
					a.agents.addAll(neighbour.area.agents);
					if (neighbour.area.modelArea instanceof Refuge)
						a.refuges.add((Refuge) neighbour.area.modelArea);
					for (AmbulanceTeam amb : neighbour.area.ambulances)
						amb.ambGroup = true;
					neighbour.area.mark = true;
					if (!neighbour.mark)
						neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances)
						if (!internal.mark) {
							if (!internal.area.mark) {
								if (!(internal.area.modelArea instanceof Refuge))
									a.cvs.addAll(internal.area.civilians);
								a.ambs.addAll(internal.area.ambulances);
								a.agents.addAll(internal.area.agents);
								if (internal.area.modelArea instanceof Refuge)
									a.refuges
											.add((Refuge) internal.area.modelArea);
								for (AmbulanceTeam amb : internal.area.ambulances)
									amb.ambGroup = true;
								internal.area.mark = true;
							}
							internal.mark = true;
							newLayer.add(internal);
						}
				}

			}
			layer = newLayer;
		}
		// amb.ambGroup = true;
		return a;
	}

	public AmbulanceTeam nearestAmb(ArrayList<Enterance> start) {
		wg.clearAreas();
		for (worldGraph.Area area : wg.areas) {
			area.ambulances = new ArrayList<AmbulanceTeam>();
		}
		// ambulances without any CV are going to add to ambareas
		for (AmbulanceTeam amb : modelAmbulanceTeams) {
			if (amb.CV == null) {
				worldGraph.Area ambArea = ((Area) model.getEntity(amb
						.getPosition())).worldGraphArea;
				ambArea.ambulances.add(amb);
			}
		}
		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		for (Enterance enterance : start) {
			if (enterance.area.ambulances.size() != 0) {
				return enterance.area.ambulances.get(0);
			}
			if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer)
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.area.ambulances.size() != 0) {
						return neighbour.area.ambulances.get(0);
					}
					neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances)
						if (!internal.mark) {
							if (internal.area.ambulances.size() != 0) {
								return internal.area.ambulances.get(0);
							}
							internal.mark = true;
							newLayer.add(internal);
						}
				}

			layer = newLayer;
		}
		return null;
	}

	int nearest = Integer.MAX_VALUE;

	public int nearestPolice(int myID, ArrayList<Integer> freePolices) {
		Entity myIDe = model.getEntityByInt(myID);
		double minpath = Double.MAX_VALUE;
		for (int p : freePolices) {
			Entity pe = model.getEntityByInt(p);
			double dist = Math.sqrt(Math.hypot(((Human) pe).getX()
					- ((Human) myIDe).getX(), ((Human) pe).getY()
					- ((Human) myIDe).getY()));
			if (minpath > dist) {
				nearest = p;
				minpath = dist;
			}
		}
		if (nearest != Integer.MAX_VALUE)
			return nearest;
		return 0;
	}

	public int nearestPolice(int x, int y, ArrayList<Integer> freePolices,
			boolean checkFirstPosition) {
		int nearest = Integer.MAX_VALUE;
		if (!checkFirstPosition) {
			double minpath = Double.MAX_VALUE;
			for (int p : freePolices) {
				Entity pe = model.getEntityByInt(p);
				double dist = Mathematic.getDistance(x, y, ((Human) pe).getX(),
						((Human) pe).getY());
				if (minpath > dist) {
					nearest = p;
					minpath = dist;
				}
			}
		}
		if (checkFirstPosition) {
			double minpath = Double.MAX_VALUE;
			for (int p : freePolices) {
				Entity pe = model.getEntityByInt(policeFirstPositions.get(p));
				double dist = Mathematic.getDistance(x, y, ((Area) pe).getX(),
						((Area) pe).getY());
				if (minpath > dist) {
					nearest = p;
					minpath = dist;
				}
			}
		}
		if (nearest != Integer.MAX_VALUE)
			return nearest;
		return -1;
	}

	byte[] lastSaidByteArray = null;

	protected void sendClearer() {
		if (this instanceof PoliceForceAgent) {
			if (time > 3) {
				if (commandHistory.get(time - 1).getAction().equals(
						StandardMessageURN.AK_CLEAR)) {
					int lastClearedTarget = commandHistory.get(time - 1).lastClearedTarget
							.getValue();
					if (!heardclearers.contains(lastClearedTarget)) {
						Entity entit = model.getEntityByInt(lastClearedTarget);
						if (entit instanceof Road) {
							worldGraph.Area area = ((Area) entit).worldGraphArea;
							boolean isItCleanEnough = true;
							int pos = entit.getID().getValue();
							for (Enterance e : area.enterances)
								if (e.internalEnterances.size() != area.enterances
										.size() - 1)
									isItCleanEnough = false;
							if (isItCleanEnough == true) {
								byte[] byteArray = radar.codeForClearer(
										allEntities.get(pos), area);
								lastSaidByteArray = new byte[byteArray.length];
								for (int i = 0; i < byteArray.length; i++) {
									lastSaidByteArray[i] = byteArray[i];
								}
								sendSpeak(time, channel, byteArray);
								if (channel != 0)
									sendSpeak(time, 0, byteArray);
							}
						}
					}
				}
			}
		}
	}

	protected void hearClearer() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_CLEAR) {
							int[] array = radar.decodeForClearer(recivedSpeak,
									model, allRanks);
							int pos = array[0];
							Entity entit = model.getEntityByInt(pos);
							Area _area = (Area) entit;
							worldGraph.Area area = _area.worldGraphArea;
							heardclearers.add(pos);
							if (!changeSet.getChangedEntities().contains(
									entit.getID())) {
								area.blockades.clear();
								area.updateAvailablePoints();
								area.updateEnterancesConnectivity(
										WorldGraph.RAD, changeSet, true);
								for (Enterance e : area.enterances) {
									e.internalEnterances.clear();
									for (Enterance f : area.enterances)
										if (f != e)
											e.internalEnterances.add(f);
								}
								for (int i = 1; i < array.length; i++)
									if (!changeSet
											.getChangedEntities()
											.contains(
													area.enterances.get(i - 1).neighbour.area.modelArea
															.getID()))
										area.enterances.get(i - 1).isItConnectedToNeighbour = area.enterances
												.get(i - 1).neighbour.isItConnectedToNeighbour = (array[i] == 1);
							}
						}
					}
				}
			}
		}
	}

	ArrayList<Integer> forSend = new ArrayList<Integer>();

	protected void centerForAmb() {
		groupingForAmb();
		// if (lastALForAmb.size() == 0)
		forSend = new ArrayList<Integer>();
		for (int i = 0; i < modelAmbulanceTeams.size(); i++) {
			forSend.add(-1);
			forSend.add(0);
		}
		for (AmbsCvs ambcv : groups) {
			String str = "";
			for (AmbulanceTeam a : ambcv.ambs)
				str += a.getID().getValue() + " ";
			String str2 = "";
			for (Civilian c : ambcv.cvs)
				str2 += c.getID().getValue() + " ";
			log("CC: " + str);
			log("CC2: " + str2);
		}
		// log("1.forSend: " + forSend);
		// for (int i = 0; i < modelAmbulanceTeams.size() * 2; i += 2) {
		// if (modelAmbulanceTeams.get(i / 2).target == -1) {
		// forSend.set(i, -1);
		// forSend.set(i + 1, 0);
		// }
		// }
		// log("2.forSend: " + forSend);
		for (AmbsCvs ambcv : groups) {
			if (ambcv.cvs.size() > 0 || ambcv.agents.size() > 0) {
				setCVsAndAgentsForAmbs(ambcv);
				setMainJobs(ambcv.state);
			}
		}
		log("3.forSend: " + forSend);
		for (int a = 0; a < forSend.size(); a++) {
			if (lastALForAmb.size() != forSend.size()
					|| forSend.get(a) != lastALForAmb.get(a)) {
				log("HA? " + " " + a);
				sendAmbMap(forSend);
				lastALForAmb = forSend;
				break;
			}
		}
	}

	protected void setCVsAndAgentsForAmbs(AmbsCvs group) {
		for (Civilian c : group.cvs) {
			int x = -1, y = -1;
			if (model.getEntity(c.getPosition()) instanceof AmbulanceTeam) {
				AmbulanceTeam ambulanceTeam = (AmbulanceTeam) model.getEntity(c
						.getPosition());
				x = ambulanceTeam.getX();
				y = ambulanceTeam.getY();
			} else {
				x = c.getX();
				y = c.getY();
			}
			Entity e = model.getEntity(c.getPosition());
			c.setNearestRefuge(group.refuges, x, y, e);
			if (c.nearestRefuge == null) {
				if (reachableRoads.size() > 0)
					c.nearestRefuge = reachableRoads.get(0).getID();
				else if (reachableBuildings.size() > 0) {
					if (reachableBuildings.get(0).getID() != c.getPosition())
						c.nearestRefuge = reachableBuildings.get(0).getID();
					else
						c.nearestRefuge = reachableBuildings.get(1).getID();
				} else
					c.nearestRefuge = modelRefuges.get(0).getID();
			}
		}
		Collections.sort(group.cvs);
		Collections.sort(group.agents);
		Collections.sort(group.ambs, IDcomparator);
		group.state = new State(group.cvs, group.agents, group.ambs, model);
		group.state.generateList();
	}

	protected void setMainJobs(State state) {
		if (state.getList() != null) {
			for (int i = state.getList().length - 1; i >= 0; i--) {
				ArrayList<Integer> IDs = new ArrayList<Integer>();
				int c = 0, s = state.getAgents().size(), pos;
				if (state.getList()[i] != null) {
					c = modelAmbulanceTeams.indexOf(state.getList()[i]) * 2;
					Human h = null;
					if (i < s)
						h = state.getAgents().get(i);
					else
						h = state.getCvs().get(i - s);
					pos = h.getPosition().getValue();
					for (Human a : Agents)
						if (a.getPosition().getValue() == pos)
							IDs.add(a.getID().getValue());
					for (Entity e : model
							.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
						Civilian cv = (Civilian) e;
						if (cv.isPositionDefined()
								&& cv.getPosition().getValue() == pos) {
							IDs.add(cv.getID().getValue());
						}
					}
					Collections.sort(IDs);
					forSend.set(c, pos);
					forSend.set(c + 1, IDs.indexOf(h.getID().getValue()));
				}
			}
		}
	}

	ArrayList<Integer> lastALForAmb = new ArrayList<Integer>();

	public Comparator<Entity> IDcomparator = new Comparator<Entity>() {
		public int compare(Entity a1, Entity a2) {
			// TODO Auto-generated method stub
			if (a1.getID().getValue() > a2.getID().getValue())
				return 1;
			if (a1.getID().getValue() < a2.getID().getValue())
				return -1;
			return 0;
		}
	};

	protected void checkRechablePolicesAvailability() {
		for (PoliceForce police : modelPoliceForces)
			if (police.isBuriednessDefined() && police.getBuriedness() > 0)
				police.hasBuriedness = true;
		if (me() instanceof Human
				&& (!(me() instanceof PoliceForce) || (((Human) me())
						.isBuriednessDefined() && ((Human) me())
						.getBuriedness() != 0))) {
			boolean b = false;
			for (PoliceForce police : modelPoliceForces)
				if (isReachable(police.getPosition()) == true
						&& police.getID().getValue() != me().getID().getValue()
						&& police.hasBuriedness == false) {
					b = true;
					break;
				}
			if (b == false)
				sendHelp();
		}
		if (longnHP.containsKey(me().getID().getValue())
				|| (wasItInLongnHP == true && counterForUnHelp < 5)) {
			wasItInLongnHP = true;
			Human me = (Human) me();
			boolean b = false;
			for (PoliceForce police : modelPoliceForces)
				if (isReachable(police.getPosition()) == true
						&& police.getID().getValue() != me().getID().getValue()
						&& police.hasBuriedness == false) {
					b = true;
					break;
				}
			if (b == true
					|| (me() instanceof PoliceForce && (me
							.isBuriednessDefined() && me.getBuriedness() == 0))) {
				unSendHelp();
			}
		}
	}

	protected void checkBriedness() {
		if (me() instanceof Human) {
			Human me = (Human) me();
			if (me.getBuriedness() > 0)
				sendHelpForAmb();
			if (longnAP.containsKey(me().getID().getValue())
					|| (wasItInLongnAP == true && counterForAmbUnHelp < 5)) {
				wasItInLongnAP = true;
				if (me.getBuriedness() == 0
						|| (me.isHPDefined() && me.getHP() == 0)) {
					unSendHelpForAmb();
				}
			}
		}
	}

	protected void sendMyPos() {
		if (time - lastSaidPositionCycle > cyclesBtweenSayingPos) {
			Human me = (Human) me();
			int pos = me.getPosition().getValue();
			byte[] byteArray = radar.code(Radar.HEADER_sAYmYpOS, allEntities
					.get(pos));
			sendSpeak(time, channel, byteArray);
			sendSpeak(time, 0, byteArray);
		}
	}

	protected void hearSOsPos() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_sAYmYpOS && noCommi == false) {
							int id = cmd.getAgentID().getValue(), pos = allRanks
									.get(radar.decode(recivedSpeak));
							addAgentPositionToModel(id, pos);
						}
					}
				}
			}
		}
	}

	// TODO

	// protected void sendWayToReuge() {
	// if (me() instanceof Human) {
	// Human me = (Human) me();

	// BitSet bits = new BitSet();
	// int start = Radar.calcWithBitSet(bits, Radar.HEADER_WAYtOrEFUGE, 0,
	// Radar.headerSize);
	// start = Radar
	// .calcWithBitSet(bits,
	// allEntities.get(me.getPosition().getValue()),
	// start, powNum);
	// ArrayList<EntityID> ways = new ArrayList<EntityID>();
	// Entity ent = model.getEntity(me.getPosition());
	// HashSet<Refuge> refuges = new HashSet<Refuge>();
	// refuges.addAll(modelRefuges);
	// me.setNearestRefuge(refuges, me.getX(), me.getY(), ent);
	// ways.addAll(wg.getMinPath(me.getPosition(), refuges, true));
	// for(int i = 0; i < ways.size(); i++)
	// {
	// if()
	// }
	// }
	// }

	protected void SendNoCommiForFire(HashSet<Building> buildings) {
		if (buildings.size() > 0) {
			BitSet bits = new BitSet();
			bits = radar.mapHeader(Radar.HEADER_BuilDingOutOFreAchNeaRRoads);
			int start = Radar.headerSize;
			int fireness = 0;
			for (Building build : buildings) {
				if (!build.isFierynessDefined())
					fireness = 0;
				else
					fireness = build.getFieryness();
				int lts = 0;
				if (build.lastTimeSeen > 0)
					lts = build.lastTimeSeen;
				start = radar.buildingCode(bits, allEntities.get(build.getID()
						.getValue()), fireness, lts, start);
			}
			bits.set(start);
			byte[] byteArray = Radar.toByteArray(bits);
			sendSpeak(time, channel, byteArray);
		}
	}

	protected void sendEmptyBuildings(Set<Building> emptyBuildings) {
		if (emptyBuildings != null && emptyBuildings.size() > 0) {
			BitSet bits = new BitSet();
			int start = 0;
			start = Radar.calcWithBitSet(bits, Radar.HEADER_EMPTYbUILDINGS,
					start, Radar.headerSize);
			for (Building b : emptyBuildings) {
				start = Radar.calcWithBitSet(bits, allEntities.get(b.getID()
						.getValue()), start, powNum);
				start = Radar.calcWithBitSet(bits, b.lastTimeVisit, start, 9);
			}
			bits.set(start);
			byte[] byteArray = Radar.toByteArray(bits);
			sendSpeak(time, 0, byteArray);
		}
	}

	protected void hearEmptyBuildings() {
		if (heard.size() > 0) {
			byte[] recivedSpeak = null;
			for (Command cmd : heard) {
				Entity entity = model.getEntityByInt(cmd.getAgentID()
						.getValue());
				if (cmd instanceof AKSpeak && entity != null
						&& !(entity instanceof Civilian)) {
					AKSpeak as = (AKSpeak) cmd;
					recivedSpeak = as.getContent();
					if (recivedSpeak.length > 0) {
						int header = radar.tellHeader(recivedSpeak);
						if (header == Radar.HEADER_EMPTYbUILDINGS) {
							ArrayList<Integer> emptyBuilds = new ArrayList<Integer>();
							emptyBuilds.addAll(radar
									.decodeForEmptyBuildings(recivedSpeak));
							for (int i = 0; i < emptyBuilds.size(); i += 2) {
								Building build = (Building) model
										.getEntityByInt(allRanks
												.get(emptyBuilds.get(i)));
								if (!changeSet.getChangedEntities().contains(
										build))
									build.lastTimeVisit = emptyBuilds
											.get(i + 1);
							}
						}
					}
				}
			}
		}
	}

	boolean haveIHeardFromRadar = false;

	protected void radarBeforeDecide() {
		radar = new Radar(powNum);
		radar.logger = logger;
		radar.sentBlockades = sentBlockades;
		counterForBusiness = 0;
		selectChannel();
		selectCenterForPolice();
		selectCenterForAmb();
		sendSubscribe(time, channel);
		if (!noCommi) {
			hearHelp();
			unHearHelp();
			hearHelpForAmb();
			unHearHelpForAmb();
			hearBuildingInformation();
			if (me() instanceof PoliceForce)
				hearPoliceMap();
			hearCVlInformation();
			hearWay();
			hearNoWay();
			hearClearer();
			hearCheckForOthersOfAmb();
			hearSOsPos();
			hearAmbMap();
			if (me() instanceof Human) {
				checkRechablePolicesAvailability();
				checkBriedness();
			}
			if (me().getID().getValue() == centerForAmb/* && time % 10 == 6 */)
				centerForAmb();
			sendBuildingInformation();
			if (me().getID().getValue() == centerForPolice)
				centerForPolice();
		}
	}

	boolean isItNeededToSay = false;

	protected void radarAfterDecide() {
		if (counterForUnLoad == 0 && lastSaidUnLoadTime == time - 1) {
			sendCheckForOthersOfAmb(0);
		} else if (counterForUnLoad >= 1)
			counterForUnLoad = -1;
		boolean isACVOnMyCat = false;
		for (StandardEntity next : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian civilian = (Civilian) next;
			if (civilian.isPositionDefined())
				if (civilian.getPosition().equals(getID())) {
					isACVOnMyCat = true;
					break;
				}
		}
		if (lastRescuedTime == time - 1 && isACVOnMyCat == false) {
			sendCheckForOthersOfAmb(0);
			isItNeededToSay = true;
			lastSaidUnLoadTime = time;
			counterForRescue = -1;
		}
		if (isItNeededToSay && lastSaidUnLoadTime == time - 1) {
			sendCheckForOthersOfAmb(0);
			isItNeededToSay = false;
		}
		if (!noCommi) {
			sendCVlInformation();
			if (time > 10) {
				checkWays();
				if (lastSaidByteArray != null && lastSaidByteArray.length != 0) {
					sendSpeak(time, channel, lastSaidByteArray);
					if (channel != 0)
						sendSpeak(time, 0, lastSaidByteArray);
					lastSaidByteArray = null;
				}
				sendClearer();
			}
			if (me() instanceof FireBrigade && noCommi == false) {
				Human me = (Human) me();
				if (!me.isBuriednessDefined() || me.getBuriedness() == 0)
					sendMyPos();
			}
		} else {
			hearHelp();
			hearHelpForAmb();
			hearCheckForOthersOfAmb();
			if (me() instanceof FireBrigade)
				hearBuildingInformation();
			hearCVlInformation();
			hearEmptyBuildings();
			if (me() instanceof Human) {
				Human me = (Human) me();
				for (PoliceForce police : modelPoliceForces)
					if (police.isBuriednessDefined()
							&& police.getBuriedness() > 0)
						police.hasBuriedness = true;
				if (me() instanceof Human
						&& (!(me() instanceof PoliceForce) || (((Human) me())
								.isBuriednessDefined() && ((Human) me())
								.getBuriedness() != 0))) {
					boolean b = false;
					for (PoliceForce police : modelPoliceForces)
						if (isReachable(police.getPosition()) == true
								&& police.getID().getValue() != me().getID()
										.getValue()
								&& police.hasBuriedness == false) {
							b = true;
							break;
						}
					if (b == false)
						sendHelp();
				}
				if (me.getBuriedness() > 0)
					sendHelpForAmb();
			}
			sendCVlInformationForNoCommi();
		}
	}

	protected int time = -1;
	protected Collection<Command> heard = null;
	protected Map<Integer, ActionCommandException> commandHistory = new HashMap<Integer, ActionCommandException>();
	public ChangeSet changeSet = null;

	public int STUCKCYCLES = 4;
	public ArrayList<AmbulanceTeam> ambulances = new ArrayList<AmbulanceTeam>(); // for
	// ambulance
	public int[] positions = new int[STUCKCYCLES];

	public HashSet<Building> buildsOfFire = new HashSet<Building>();
	public ArrayList<Building> emptyB = new ArrayList<Building>();
	public int manlastDeadTime = 0;

	protected void think(int time, ChangeSet changes, Collection<Command> heard) {
		try {
			cycleTimer.setTime(0);
			log("Think started");
			this.time = time;
			this.heard = heard;
			this.changeSet = changes;
			allPpcs = new ArrayList<PolicePC>();
			for (EntityID forroyechangset : changes.getChangedEntities()) {
				Entity entity = model.getEntity(forroyechangset);
				if (entity instanceof Area) {
					if (entity instanceof Building
							&& ((Building) entity).lastTimeSeen == LTSDORI)
						buildsOfFire.add(((Building) entity));
					((Area) entity).lastTimeSeen = time;
				}
			}

			log("before world graph update");
			wg.logger = logger;
			wg.update(changes);
			log("after world graph update");
			if (me() instanceof Human) {
				wg.update(time, this, ((Human) me()).getPosition(),
						((Human) me()).getX(), ((Human) me()).getY());
				((Area) model.getEntity(((Human) me()).getPosition())).lastTimeVisit = time;
			} else
				wg.update(time, this, null, -1, -1);
			for (EntityID e : changeSet.getChangedEntities()) {
				Entity entity = model.getEntity(e);
				if (entity instanceof Human) {
					Human h = (Human) entity;
					h.setDeadtime(time);
				}
			}
			logger.logTime(time);
			if (me() instanceof Human &&
			// TODO nillluuuuuu!!!!! sar be Dvar mikoBm!!!!!no commi bash!
					// avarin!
					noCommi) {
				log("MyPosition: " + ((Human) me()).getPosition());

				if ((model.getEntity(((Human) me()).getPosition())) instanceof Building)
					for (EntityID entityid : changeSet.getChangedEntities()) {
						Entity NTT = model.getEntity(entityid);
						if (NTT instanceof Civilian) {
							break;
						} else
							emptyB.add((Building) model
									.getEntity(((Human) me()).getPosition()));
					}
				Set<Building> emptyB2 = new HashSet<Building>();
				int size = emptyB.size();
				if (size > 20)
					size = 20;
				for (int i = 1; i <= size; i++)
					emptyB2.add(emptyB.get(emptyB.size() - i));
				sendEmptyBuildings(emptyB2);
			}
			try {
				try {
					HashSet<Building> buildHayeSayVajeb = new HashSet<Building>();
					ArrayList<Building> buildsKol = new ArrayList<Building>();
					for (Building building : modelBuildings)
						if (building.isOnFire())
							buildsKol.add(building);
					Collections.sort(buildsKol, lTSComparator);
					for (Building building : buildsKol) {
						int u = 0;
						if (u >= 20)
							break;
						buildHayeSayVajeb.add(building);
						u++;
					}
					if (buildHayeSayVajeb.size() != 0 && noCommi)
						SendNoCommiForFire(buildHayeSayVajeb);

					radarBeforeDecide();
				} catch (Exception e) {
					e.printStackTrace();
					if (logger.getWriter() != null) {
						System.out.println("me : " + me().getID().getValue());
						e.printStackTrace(logger.getWriter());
					}
				}

				if (me() instanceof Human)
					wg.updateReachableAreas(wg.myEnterances);

				if (me() instanceof Human && time > STUCKCYCLES) {
					positions[time % STUCKCYCLES] = ((Human) me())
							.getPosition().getValue();
					if (shouldBaseDecideForRandomWalk && time > STUCKCYCLES) {
						boolean stucked = true;
						for (int i = 0; i < STUCKCYCLES - 1; i++)
							if (positions[i] != positions[i + 1]
									|| !commandHistory.get(time - (i + 1))
											.getAction().equals(
													StandardMessageURN.AK_MOVE)) {
								stucked = false;

								break;
							}

						if (stucked
								&& !(me() instanceof FireBrigade && reachableAreas
										.size() == 1)) {
							log("RandomWalk Executed for some problems");
							randomWalk();
						}
					}
				}

				if (me() instanceof Human && !(me() instanceof PoliceForce))
					setMyZone(true);

				ActionCommandException ace = null;
				shouldBaseDecideForRandomWalk = true;
				try {
					if (me() instanceof Human) {
						Human man = (Human) me();
						man.setDeadtime(time);
						log("deadTime: " + man.deadtime);
						if (movingToRefuge && reachableRefuges.size() > 0
								&& manlastDeadTime <= man.deadtime) {
							shouldBaseDecideForRandomWalk = false;
							move(reachableRefuges.get(0).getID());
						}
						if (man.deadtime < 300 && reachableRefuges.size() > 0) {
							int dist = wg.getMinPath(man.getPosition(),
									reachableRefuges.get(0).getID(), true)
									.size();
							log("dist: " + dist);
							// int dist = (int) ((Mathematic.getDistance(man
							// .getX(), man.getY(), reachableRefuges
							// .get(0).getX(), reachableRefuges.get(0)
							// .getY())) / 1000);
							if (time > man.deadtime - dist) {
								movingToRefuge = true;
								log("move to refuge");
								manlastDeadTime = man.deadtime;
								shouldBaseDecideForRandomWalk = false;
								move(reachableRefuges.get(0).getID());
							}
						}
					}
					decide();
				} catch (ActionCommandException ac) {
					ace = ac;
				}
				if (ace == null)
					rest();
				throw ace;
			} catch (ActionCommandException ace) {
				log("Command: " + ace.getAction() + " Sent");
				commandHistory.put(time, ace);
			}
		} catch (Exception e) {
			e.printStackTrace();
			if (logger.getWriter() != null)
				e.printStackTrace(logger.getWriter());
			sendRest(time);
			commandHistory.put(time, new ActionCommandException(
					StandardMessageURN.AK_REST));
		}
		try {
			radarAfterDecide();
		} catch (Exception e) {
			e.printStackTrace();
			if (logger.getWriter() != null) {
				System.out.println("me : " + me().getID().getValue());
				e.printStackTrace(logger.getWriter());
			}
		}
		logger.logEndOfCycle();
	}

	public Comparator<Building> lTSComparator = new Comparator<Building>() {
		public int compare(Building a, Building b) {
			if (a.lastTimeSeen > b.lastTimeSeen)
				return -1;
			if (a.lastTimeSeen < b.lastTimeSeen)
				return 1;
			return 0;
		}
	};

	protected void rest() throws ActionCommandException {
		sendRest(time);
		throw new ActionCommandException(StandardMessageURN.AK_REST);
	}

	protected void extinguish(EntityID target, int power)
			throws ActionCommandException {
		for (Building building : ((Building) model.getEntity(target)).nearBuildings)
			if (!building.isFierynessDefined() || building.getFieryness() == 0) {
				building.lastTimeSeen = LTSDORI;
			}
		sendExtinguish(time, target, power);
		throw new ActionCommandException(StandardMessageURN.AK_EXTINGUISH);
	}

	protected void move(EntityID target) throws ActionCommandException {
		if (me() instanceof PoliceForce)
			move(target, false);
		else
			move(target, true);
	}

	protected void move(EntityID target, boolean checkBlockades)
			throws ActionCommandException {
		if (!(me() instanceof Human))
			return;

		Human m = (Human) me();
		ArrayList<EntityID> way = wg.getMinPath(m.getPosition(), target,
				checkBlockades);
		if (way.size() > 0
				&& way.get(0).getValue() == m.getPosition().getValue())
			way.remove(0);
		move(way);
	}

	protected void move(List<EntityID> path) throws ActionCommandException {
		sendMove(time, path);
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	protected void move(List<EntityID> path, int x, int y)
			throws ActionCommandException {
		sendMove(time, path, x, y);
		throw new ActionCommandException(StandardMessageURN.AK_MOVE);
	}

	protected void clear(EntityID target) throws ActionCommandException {
		sendClear(time, target);
		ActionCommandException ace = new ActionCommandException(
				StandardMessageURN.AK_CLEAR);
		ace.lastClearedTarget = ((Blockade) model.getEntity(target))
				.getPosition();
		throw ace;

	}

	protected void rescue(EntityID target) throws ActionCommandException {
		lastRescuedTime = time;
		lastRescuedObject = model.getEntity(target).getID();
		((AmbulanceTeam) model.getEntityByInt(me().getID().getValue())).target = target
				.getValue();
		((Human) model.getEntity(target)).owner = me().getID().getValue();
		if (noCommi) {
			sendCheckForOthersForNoCommi(target.getValue());
		} else
			sendCheckForOthersOfAmb(target.getValue());
		sendRescue(time, target);
		throw new ActionCommandException(StandardMessageURN.AK_RESCUE);
	}

	protected void load(EntityID target) throws ActionCommandException {
		lastRescuedTime = time;
		((AmbulanceTeam) model.getEntityByInt(me().getID().getValue())).target = target
				.getValue();
		((Human) model.getEntity(target)).owner = me().getID().getValue();
		sendCheckForOthersOfAmb(target.getValue());
		sendLoad(time, target);
		throw new ActionCommandException(StandardMessageURN.AK_LOAD);
	}

	protected void unLoad() throws ActionCommandException {
		counterForRescue = -1;
		if (noCommi)
			sendCheckForOthersForNoCommi(0);
		else
			sendCheckForOthersOfAmb(0);
		((AmbulanceTeam) model.getEntityByInt(me().getID().getValue())).target = -1;
		sendUnload(time);
		throw new ActionCommandException(StandardMessageURN.AK_UNLOAD);
	}

	// protected boolean isReachable(Enterance enterance) {
	// return (reachableEnterances.get(enterance.index) != null);
	// }

	protected boolean isReachable(EntityID target) {
		StandardEntity se = model.getEntity(target);
		if (se == null || !(se instanceof Area))
			return false;
		Area area = (Area) se;
		return area.worldGraphArea.isReachable;
		// return (reachableAreas.get(target.getValue()) != null);
	}

	protected void log(String msg) {
		// logger.log(msg);
	}

	public int getTime() {
		return time;
	}

	protected boolean isInChangeSet(EntityID entityID) {
		if (changeSet == null)
			return false;

		return changeSet.getChangedEntities().contains(entityID);
	}

	public StandardWorldModel getModel() {
		return model;
	}

	private static final int RANDOM_WALK_LENGTH = 50;
	private Map<EntityID, Set<EntityID>> neighbours;

	protected void randomWalk() throws ActionCommandException {
		List<EntityID> result = new ArrayList<EntityID>(RANDOM_WALK_LENGTH);
		Set<EntityID> seen = new HashSet<EntityID>();
		EntityID current = ((Human) me()).getPosition();
		for (int i = 0; i < RANDOM_WALK_LENGTH; ++i) {
			result.add(current);
			seen.add(current);
			List<EntityID> possible = new ArrayList<EntityID>(neighbours
					.get(current));
			Collections.shuffle(possible, new Random());
			boolean found = false;
			for (EntityID next : possible) {
				if (seen.contains(next)) {
					continue;
				}
				current = next;
				found = true;
				break;
			}
			if (!found) {
				// We reached a dead-end.
				break;
			}
		}

		move(result);
	}

	protected Zone searchZone = null;

	private ArrayList<Integer> IDs = new ArrayList<Integer>();

	int numberOfZones = 5;
	int sizeOfNotManfiyeYekPFs = 0;

	protected void setMyZone(boolean chooseMyself) {
		if (policeFirstPositions == null) {
			policeFirstPositions = new HashMap<Integer, Integer>();
			for (PoliceForce pf : modelPoliceForces)
				policeFirstPositions.put(pf.getID().getValue(), ((Human) pf)
						.getPosition().getValue());
		}
		lastIDs = new ArrayList<Integer>();
		lastIDs.addAll(IDs);
		IDs = new ArrayList<Integer>();
		if (this instanceof PoliceForceAgent) {
			for (PoliceForce policeForce : modelPoliceForces)
				if (!policeForce.hasBuriedness || noCommi)
					IDs.add(policeForce.getID().getValue());
			numberOfZones = Math.min(5, (IDs.size() + 1) / 2);
		}
		if (this instanceof FireBrigadeAgent) {
			for (FireBrigade fireBrigade : modelFireBrigades)
				if (!fireBrigade.hasBuriedness || noCommi)
					IDs.add(fireBrigade.getID().getValue());
			numberOfZones = Math.min(5, IDs.size());
		}
		if (this instanceof AmbulanceTeamAgent) {
			for (AmbulanceTeam ambulanceTeam : modelAmbulanceTeams)
				if (noCommi
						|| (!ambulanceTeam.hasBuriedness && (!ambulanceTeam
								.isHPDefined() || ambulanceTeam.getHP() != 0)))
					IDs.add(ambulanceTeam.getID().getValue());
			numberOfZones = Math.min(5, IDs.size());
		}

		if (!(me() instanceof PoliceForce) && lastIDs.size() == IDs.size())
			return;

		if (me() instanceof Human) {
			Human m = (Human) me();
			if (m.hasBuriedness || (m.isHPDefined() && m.getHP() == 0)) {
				mySubZone = null;
				return;
			}
		}
		// zoning the map
		zoning(numberOfZones);
		// my index =chandomin agent 0,1,2,...
		Collections.sort(IDs);
		int myIndex = IDs.indexOf(me().getID().getValue());

		// subZoning or-and policeZoning my zone...
		if (this instanceof PoliceForceAgent) {
			for (Zone z : zones)
				policeZoning(z.zoneNumber, z.policeZones.length);
			// setting my OWN police zone
			if (chooseMyself) {
				ArrayList<Integer> freePolices = new ArrayList<Integer>();
				freePolices.addAll(IDs);
				done: while (true)
					for (Zone z : zones)
						for (PoliceZone pz : z.policeZones) {
							if (freePolices.size() == 0)
								break done;
							int nearest = nearestPolice(pz.center.getX(),
									pz.center.getY(), freePolices, true);
							if (nearest == me().getID().getValue()) {
								myPoliceZone = pz;
								myZoneNumber = z.zoneNumber;
							}
							pz.policesDedicatedto.add(nearest);
							freePolices.remove(freePolices.indexOf(nearest));
						}
				for (Zone z : zones)
					for (PoliceZone pz : z.policeZones) {
						pz.setSubZonesForPolice(pz.policesDedicatedto.size());
						subZoningForPolice(z.zoneNumber, pz.policeZoneNumber,
								pz.subZones.length);
					}

				// setting my OWN subZone for polices
				int nearest = 0;
				freePolices.addAll(myPoliceZone.policesDedicatedto);
				for (SubZone sz : myPoliceZone.subZones) {
					nearest = nearestPolice(sz.center.getX(), sz.center.getY(),
							freePolices, true);
					if (nearest == me().getID().getValue()) {
						mySubZone = sz;
						mySubZoneNumber = sz.subZoneNumber;
					}
					freePolices.remove(freePolices.indexOf(nearest));
				}
			} else {
				sizeOfNotManfiyeYekPFs = findIDsOfFreePolices(
						policeDecisionHeard1).size();
				setPolicesDedicatedTo();
				for (Zone z : zones) {
					for (PoliceZone pz : z.policeZones) {
						pz.setSubZonesForPolice(pz.dedicatedTo);
						subZoningForPolice(z.zoneNumber, pz.policeZoneNumber,
								pz.subZones.length);
					}
				}

			}
		} else {
			for (Zone z : zones)
				subZoning(z.zoneNumber, z.subZones.length);
			// setting myOWN sub zone:
			int j = -1;
			done: while (true) {
				j++;
				for (int i = 0; i < numberOfZones; i++) {
					mySubZone = zones[i % numberOfZones].subZones[j];
					myZone1 = zones[i % numberOfZones];
					if (numberOfZones * (j) + (i) == (myIndex))
						break done;
				}
			}
		}
	}

	public void setPolicesDedicatedTo() {
		int counter = 0;
		done: while (true) {
			for (Zone z : zones)
				for (PoliceZone pz : z.policeZones) {
					if (counter == sizeOfNotManfiyeYekPFs)
						break done;
					counter++;
					pz.dedicatedTo++;
				}
		}
	}

	private void setCenterOfMap() {
		int minX = Integer.MAX_VALUE, maxX = Integer.MIN_VALUE;
		int minY = Integer.MAX_VALUE, maxY = Integer.MIN_VALUE;
		for (Area area : modelAreas) {
			if (area.getX() > maxX)
				maxX = area.getX();
			if (area.getX() < minX)
				minX = area.getX();
			if (area.getY() > maxY)
				maxY = area.getY();
			if (area.getY() < minY)
				minY = area.getY();
		}
		centerOfMap = new Point((minX + maxX) / 2, (minY + maxY) / 2);
	}

	Zone zones[];
	ArrayList<SubZone> subzones;

	public Point centerOfMap = null;

	private void zoning(int zonesCount) {
		zones = new Zone[zonesCount];
		// pCount = tedade hadeghal agent ha baraye 1 zone
		int agentsCount = IDs.size();
		int pCount = agentsCount / zonesCount;
		// setting subzones...
		if (this instanceof PoliceForceAgent) {
			// setting policeZones...
			pCount = Math.min(2, pCount);
			for (int i = 0; i < zones.length; i++) {
				zones[i] = new Zone(i);
				if (pCount == 2) {
					zones[i].setPoliceZones(2);
					zones[i].policeZones[0] = new PoliceZone(0);
					zones[i].policeZones[1] = new PoliceZone(1);
				} else if (i < agentsCount % zonesCount) {
					zones[i].setPoliceZones(pCount + 1);
					for (int j = 0; j <= pCount; j++)
						zones[i].policeZones[j] = new PoliceZone(j);
				} else {
					zones[i].setPoliceZones(pCount);
					for (int v = 0; v < pCount; v++)
						zones[i].policeZones[v] = new PoliceZone(v);
				}
			}
		} else {
			// setting subZones...
			for (int i = 0; i < zones.length; i++) {
				zones[i] = new Zone(i);
				if (i < agentsCount % zonesCount) {

					zones[i].setSubZones(pCount + 1);
					for (int j = 0; j <= pCount; j++)
						zones[i].subZones[j] = new SubZone(j);
				} else {
					zones[i].setSubZones(pCount);
					for (int v = 0; v < pCount; v++)
						zones[i].subZones[v] = new SubZone(v);
				}
			}
		}

		for (Area area : modelAreas) {
			float theta = (new Line(centerOfMap, new Point(area.getX(), area
					.getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta < 0)
				theta += 360;

			int zoneNumber = (int) (theta / (360.0 / zonesCount));
			area.worldGraphArea.ZoneNumberForPolice = zoneNumber;
			zones[zoneNumber].areas.add(area);
			if (area instanceof Refuge)
				zones[zoneNumber].refuges.add((Refuge) area);
			else if (area instanceof Building)
				zones[zoneNumber].buildings.add((Building) area);
			else if (area instanceof Road)
				zones[zoneNumber].roads.add((Road) area);
		}
		for (Zone z : zones) {
			int minX = Integer.MAX_VALUE, maxX = Integer.MIN_VALUE;
			int minY = Integer.MAX_VALUE, maxY = Integer.MIN_VALUE;
			for (Area area : z.areas) {
				if (area.getX() > maxX)
					maxX = area.getX();
				if (area.getX() < minX)
					minX = area.getX();
				if (area.getY() > maxY)
					maxY = area.getY();
				if (area.getY() < minY)
					minY = area.getY();
			}
			z.center = new Point((minX + maxX) / 2, (minY + maxY) / 2);
		}

	}

	protected void subZoning(int zoneNumber, int subZonesCount) {
		for (Area area : zones[zoneNumber].areas) {
			float theta = (new Line(zones[zoneNumber].center, new Point(area
					.getX(), area.getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta < 0)
				theta += 360;

			int subZoneNumber = (int) (theta / (360.0 / subZonesCount));
			zones[zoneNumber].subZones[subZoneNumber].areas.add(area);
			if (area instanceof Refuge)
				zones[zoneNumber].subZones[subZoneNumber].refuges
						.add((Refuge) area);
			else if (area instanceof Building)
				zones[zoneNumber].subZones[subZoneNumber].buildings
						.add((Building) area);
			else if (area instanceof Road)
				zones[zoneNumber].subZones[subZoneNumber].roads
						.add((Road) area);

		}
		for (SubZone sz : zones[zoneNumber].subZones) {
			int minX = Integer.MAX_VALUE;
			int maxX = Integer.MIN_VALUE;
			int minY = Integer.MAX_VALUE;
			int maxY = Integer.MIN_VALUE;
			for (Area area : sz.areas) {
				if (area.getX() > maxX)
					maxX = area.getX();
				if (area.getX() < minX)
					minX = area.getX();
				if (area.getY() > maxY)
					maxY = area.getY();
				if (area.getY() < minY)
					minY = area.getY();
			}
			sz.center = new Point((minX + maxX) / 2, (minY + maxY) / 2);
		}
	}

	public void subZoningForPolice(int zoneNumber, int policeZoneNumber,
			int subZonesCount) {
		for (Area area : zones[zoneNumber].policeZones[policeZoneNumber].areas) {
			float theta = (new Line(
					zones[zoneNumber].policeZones[policeZoneNumber].center,
					new Point(area.getX(), area.getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta < 0)
				theta += 360;
			int subZoneNumber = (int) (theta / (360.0 / subZonesCount));
			area.worldGraphArea.subZoneNumberForPolice = subZoneNumber;
			zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber].areas
					.add(area);
			if (area instanceof Refuge)
				zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber].refuges
						.add((Refuge) area);
			else if (area instanceof Building)
				zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber].buildings
						.add((Building) area);
			else if (area instanceof Road)
				zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber].roads
						.add((Road) area);
			if (area instanceof Building) {
				Building b = (Building) area;
				SubZone sz = zones[zoneNumber].policeZones[policeZoneNumber].subZones[subZoneNumber];
				if (b.isFierynessDefined() && b.getFieryness() > 0
						&& !sz.isIgnited) {
					sz.isIgnited = true;
					sz.priority += 5;
				}
			}
		}
		for (SubZone sz : zones[zoneNumber].policeZones[policeZoneNumber].subZones) {
			int minX = Integer.MAX_VALUE;
			int maxX = Integer.MIN_VALUE;
			int minY = Integer.MAX_VALUE;
			int maxY = Integer.MIN_VALUE;
			for (Area area : sz.areas) {
				if (area.getX() > maxX)
					maxX = area.getX();
				if (area.getX() < minX)
					minX = area.getX();
				if (area.getY() > maxY)
					maxY = area.getY();
				if (area.getY() < minY)
					minY = area.getY();
			}
			sz.center = new Point((minX + maxX) / 2, (minY + maxY) / 2);
		}

	}

	public SubZone getSubzoneForAmb(int myID) {
		ArrayList<Integer> ambulances = new ArrayList<Integer>();
		for (AmbulanceTeam amb : modelAmbulanceTeams)
			ambulances.add(amb.getID().getValue());
		Collections.sort(ambulances);
		int myIndex = 0;
		for (int iD : ambulances) {
			if (myID == iD)
				break;
			myIndex++;
		}

		int j = -1, n = -1;
		while (true) {
			if (zones.length * (j) + (n) == (myIndex))
				break;
			j++;
			for (int i = 0; i < zones.length; i++) {
				n = i;
				if (numberOfZones * (j) + (n) == (myIndex))
					return zones[i % zones.length].subZones[j];
			}
		}

		return null;

	}

	public void policeZoning(int zoneNumber, int policeZonesCount) {
		for (Area area : zones[zoneNumber].areas) {
			float theta = (new Line(zones[zoneNumber].center, new Point(area
					.getX(), area.getY()))).getTheta()
					* Degree.RAD2DEG;
			if (theta < 0)
				theta += 360;

			int policeZoneNumber = (int) (theta / (360.0 / policeZonesCount));
			area.worldGraphArea.policeZoneNumber = policeZoneNumber;
			zones[zoneNumber].policeZones[policeZoneNumber].areas.add(area);
			if (area instanceof Refuge)
				zones[zoneNumber].policeZones[policeZoneNumber].refuges
						.add((Refuge) area);
			else if (area instanceof Building)
				zones[zoneNumber].policeZones[policeZoneNumber].buildings
						.add((Building) area);
			else if (area instanceof Road)
				zones[zoneNumber].policeZones[policeZoneNumber].roads
						.add((Road) area);
		}
		for (PoliceZone pz : zones[zoneNumber].policeZones) {
			int minX = Integer.MAX_VALUE;
			int maxX = Integer.MIN_VALUE;
			int minY = Integer.MAX_VALUE;
			int maxY = Integer.MIN_VALUE;
			for (Area area : pz.areas) {
				if (area.getX() > maxX)
					maxX = area.getX();
				if (area.getX() < minX)
					minX = area.getX();
				if (area.getY() > maxY)
					maxY = area.getY();
				if (area.getY() < minY)
					minY = area.getY();
			}
			pz.center = new Point((minX + maxX) / 2, (minY + maxY) / 2);
		}
	}

	public void findMySubzone(ArrayList<Integer> policeDesicion2) {
		int myIndex = -1;
		ArrayList<Integer> allPoliceIDs = new ArrayList<Integer>();
		for (PoliceForce p : modelPoliceForces)
			allPoliceIDs.add(p.getID().getValue());
		myIndex = allPoliceIDs.indexOf(me().getID().getValue());
		if (myIndex == -1 || policeDesicion2 == null
				|| policeDesicion2.size() == 0) {
			mySubZone = null;
			myPoliceZone = null;
			myZoneNumber = -1;
			return;
		}
		mySubZoneID = policeDesicion2.get(myIndex);
		int counter = -1;

		done: for (Zone z : zones)
			for (PoliceZone pz : z.policeZones)
				for (SubZone sz : pz.subZones) {
					counter++;
					if (counter == mySubZoneID) {
						mySubZone = sz;
						mySubZone.subZoneID = mySubZoneID;
						myPoliceZone = pz;
						myZoneNumber = z.zoneNumber;
						break done;
					}
				}
	}

	public boolean sameAL(ArrayList<Integer> decision1,
			ArrayList<Integer> decision2) {
		if (decision1 == null || decision1.size() == 0) {
			if (decision2 == null || decision2.size() == 0) {
				return true;
			}
			return false;
		}
		for (int i = 0; i < decision1.size(); i++) {
			if (!decision1.get(i).equals(decision2.get(i))) {
				return false;
			}
		}
		return true;
	}

	public boolean needNewZoning(ArrayList<Integer> decision1,
			ArrayList<Integer> decision2) {
		int deadAgents1 = 0, deadAgents2 = 0;
		for (int a : decision1)
			if (a == -1)
				deadAgents1++;
		for (int a : decision2)
			if (a == -1)
				deadAgents2++;
		if (deadAgents1 != deadAgents2)
			return true;
		return false;
	}

	public ArrayList<Integer> findIDsOfFreePolices(ArrayList<Integer> decision) {
		ArrayList<Integer> freePFs = new ArrayList<Integer>();
		for (int PFindex = 0; PFindex < decision.size(); PFindex++) {
			if (decision.get(PFindex) != -1)
				freePFs.add(modelPoliceForces.get(PFindex).getID().getValue());
		}
		return freePFs;
	}

	protected abstract void decide() throws ActionCommandException;
}
