package agent;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.Map.Entry;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import worldGraph.Enterance;

public class AmbulanceTeamAgent extends Agent<AmbulanceTeam> {
	private ArrayList<Integer> jobs = null;
	protected ArrayList<Integer> emptyBuildings = new ArrayList<Integer>();
	private Civilian civilianOnBoard = null;

	protected void postConnect() {
		super.postConnect();
	}

	Human targetHuman = null;

	private void hearJob() throws ActionCommandException {
		if (ambDecisionHeard == null || ambDecisionHeard.size() == 0)
			return;
		if (ambDecisionHeard != null)
			jobs = ambDecisionHeard;
		int myIndex = modelAmbulanceTeams.indexOf(me());
		int target = jobs.get(myIndex * 2);

		if (target != -1)
			System.out.println("ID: " + me().getID().getValue() + " D: "
					+ target);
		if (target == -1 || emptyBuildings.contains(target)
				|| !isReachable(model.getEntityByInt(target).getID()))
			return;

		Building b = (Building) model.getEntityByInt(target);
		if (me().getPosition().getValue() != target && b.isOnFire() == false)
			move(((Entity) model.getEntityByInt(target)).getID());
		if (b.isOnFire() == true)
			return;
		Collections.sort(IDs);
		Human h = (Human) model.getEntityByInt(IDs.get(Math.min(IDs.size() - 1,
				jobs.get(myIndex * 2 + 1))));
		if (h instanceof Civilian) {
			targetHuman = h;
			rescue((Civilian) targetHuman);
		} else {
			targetHuman = (Human) h;
			rescue(targetHuman.getID());
		}
	}

		private void selectCV() throws ActionCommandException {
		log("5.selectCV");
		ArrayList<Civilian> civilians = new ArrayList<Civilian>();
		ArrayList<Civilian> cvs = new ArrayList<Civilian>();
		for (Entity entity : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian cv = (Civilian) entity;
			Entity e = model.getEntity(cv.getPosition());
			if (cv.owner == -1
					&& e instanceof Building
					&& !(e instanceof Refuge)
					&& cv.deadtime >= time + cv.getBuriedness() + 10
					&& cv.deadtime != 500
					&& cv.isPositionDefined()
					&& isReachable(cv.getPosition())
					&& ((cv.isHPDefined() && cv.getHP() > 0) || !cv
							.isHPDefined())) {
				Building b = (Building) e;
				if (b.isOnFire() == false
						&& !(emptyBuildings.contains(b.getID().getValue()))) {
					civilians.add(cv);
					if (cv.deadtime <= 250)
						cvs.add(cv);
				}
			}
		}
		ArrayList<Civilian> CVs = new ArrayList<Civilian>();
		if (cvs.size() == 0 && civilians.size() > 0) {
			Collections.sort(civilians);
			for (Civilian civilian : civilians)
				if (civilians.indexOf(civilian) < 5)
					CVs.add(civilian);
		} else if (cvs.size() > 0) {
			Collections.sort(cvs);
			for (Civilian civilian : cvs)
				if (cvs.indexOf(civilian) < 5)
					CVs.add(civilian);
		}
		if (CVs.size() > 0) {
			Collections.sort(CVs, CVcomparator);
			targetHuman = CVs.get(0);
		}
		if (targetHuman != null)
			rescue((Civilian) targetHuman);
	}

	public Comparator<Entity> CVcomparator = new Comparator<Entity>() {
		public int compare(Entity a1, Entity a2) {
			Civilian cv1 = (Civilian) a1;
			Civilian cv2 = (Civilian) a2;
			double d1 = Math.hypot(me().getX() - cv1.getX(), me().getY()
					- cv1.getY());
			double d2 = Math.hypot(me().getX() - cv2.getX(), me().getY()
					- cv2.getY());
			if (d1 > d2)
				return 1;
			if (d2 > d1)
				return -1;
			return 0;
		}
	};

	private void checkHelp() throws ActionCommandException {
		TreeMap<Integer, Integer> myAP = new TreeMap<Integer, Integer>(
				new Comparator<Integer>() {
					public int compare(Integer o1, Integer o2) {
						int a1 = longnAP.get(o1);
						int a2 = longnAP.get(o2);
						Area area1 = (Area) model.getEntityByInt(a1);
						Area area2 = (Area) model.getEntityByInt(a2);
						int dist1 = (int) Math.hypot(
								area1.getX() - me().getX(), area1.getY()
										- me().getY());
						int dist2 = (int) Math.hypot(
								area2.getX() - me().getX(), area2.getY()
										- me().getY());
						if (dist1 > dist2)
							return 1;
						if (dist2 < dist1)
							return -1;
						return 1;
					}
				});
		for (Entry<Integer, Integer> e : longnAP.entrySet()) {
			Human h = (Human) model.getEntityByInt(e.getKey());
			Building b = (Building) model.getEntityByInt(e.getValue());
			if (h.owner == -1 && h.deadtime > time && b.isOnFire() == false)
				myAP.put(e.getKey(), e.getValue());
		}

		for (Entry<Integer, Integer> m : myAP.entrySet()) {
			Human target = (Human) model.getEntityByInt(m.getKey());
			EntityID pos = model.getEntityByInt(m.getValue()).getID();
			if (!isReachable(pos))
				continue;
			if ((target.isBuriednessDefined() && target.getBuriedness() == 0)
					|| (target.isHPDefined() && target.getHP() == 0)) {
				longnAP.remove(m.getKey());
				continue;
			}

			targetHuman = target;
			checkTarget();
		}
	}

	private void setCivilianOnBoard() {
		civilianOnBoard = null;
		for (StandardEntity next : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian civilian = (Civilian) next;
			if (civilian.isPositionDefined()
					&& civilian.getPosition().equals(getID())) {
				civilianOnBoard = civilian;
				break;
			}
		}
	}

	private void rescue(Civilian civilian) throws ActionCommandException {
		HashSet<Refuge> refuges = new HashSet<Refuge>();
		refuges.addAll(reachableRefuges);
		int x = -1, y = -1;
		if (model.getEntity(civilian.getPosition()) instanceof AmbulanceTeam) {
			AmbulanceTeam ambulanceTeam = (AmbulanceTeam) model
					.getEntity(civilian.getPosition());
			x = ambulanceTeam.getX();
			y = ambulanceTeam.getY();
		} else {
			x = civilian.getX();
			y = civilian.getY();
		}
		Entity e = model.getEntity(civilian.getPosition());
		civilian.setNearestRefuge(refuges, x, y, e);
		if (model.getEntity(civilian.getPosition()) instanceof Area) {
			if (me().getPosition().getValue() != civilian.getPosition()
					.getValue())
				move(civilian.getPosition());
			Entity pos = model.getEntity(civilian.getPosition());
			if (isInChangeSet(civilian.getID()) && !(pos instanceof Refuge)) {
				if (civilian.getBuriedness() == 0)
					load(civilian.getID());
				rescue(civilian.getID());
			} else if (!(isInChangeSet(civilian.getID())))
				return;
		}
		EntityID refugeID = civilian.nearestRefuge;
		if (refugeID == null) {
			if (modelRefuges.size() > 0) {
				if (reachableBuildings.size() >= modelBuildings.size() - 10)
					noBlockade();
				search();
			} else if (reachableRoads.size() > 0)
				refugeID = reachableRoads.get(0).getID();
			else if (reachableBuildings.size() > 0) {
				int i = 0;
				if (reachableBuildings.get(0).getID().getValue() == me()
						.getPosition().getValue())
					i = 1;
				refugeID = reachableBuildings.get(i).getID();
			}
		}

		if (refugeID == null)
			rest();
		if (me().getPosition().getValue() != refugeID.getValue())
			move(refugeID);
		if (civilianOnBoard != null
				&& me().getPosition().getValue() == refugeID.getValue()) {
			targetHuman = null;
			unLoad();
		}
	}

	public EntityID target = null;

	protected void noCommi() throws ActionCommandException {
		log("nocommi");
		selectCV();
		log("afterselectCV");
		if (reachableBuildings.size() >= modelBuildings.size() - 10) {
			log("search noblockade");
			noBlockade();
		}
		log("search");
		search();
	}

	private ArrayList<Integer> IDs = new ArrayList<Integer>();

	private void checkEmptyBuildings() {
		IDs = new ArrayList<Integer>();
		for (EntityID entityID : changeSet.getChangedEntities()) {
			Entity e = model.getEntity(entityID);
			if (e instanceof Human) {
				Human h = (Human) e;
				if (h.getPosition().getValue() == me().getPosition().getValue()
						&& (h.owner == -1 || h.owner == me().getID().getValue())
						&& ((h.isBuriednessDefined() && h.getBuriedness() > 0) || h instanceof Civilian)
						&& ((h.isHPDefined() && h.getHP() > 0) || !h
								.isHPDefined()))
					IDs.add(h.getID().getValue());
			}
		}
		if (IDs.size() == 0) {
			emptyBuildings.add(me().getPosition().getValue());
			return;
		}
	}

	private void checkCivilianOnBoard() throws ActionCommandException {
		setCivilianOnBoard();
		if (civilianOnBoard != null)
			rescue(civilianOnBoard);
	}

	private void checkTarget() throws ActionCommandException {
		if (targetHuman != null) {
			Entity e = model.getEntity(targetHuman.getPosition());
			boolean isOnFire = false;
			if (e instanceof Building) {
				Building b = (Building) e;
				if (b.isOnFire())
					isOnFire = true;
			}
			if ((targetHuman.owner != -1 && targetHuman.owner != me().getID()
					.getValue())
					|| (targetHuman.isHPDefined() && targetHuman.getHP() == 0)
					|| (e instanceof Building && (emptyBuildings.contains(e
							.getID().getValue())
							|| !isReachable(e.getID()) || e instanceof Refuge))
					|| isOnFire == true
					|| (me().getPosition().getValue() == targetHuman
							.getPosition().getValue() && !isInChangeSet(targetHuman
							.getID()))
					|| (!(e instanceof Area) && targetHuman.getPosition()
							.getValue() != me().getID().getValue())
					|| e instanceof Road) {
				log("targetHuman == null");
				targetHuman = null;
			}
		}

		if (targetHuman != null) {
			log("targetHuman: " + targetHuman);
			log("emptyBuildings: " + emptyBuildings);
			if (targetHuman instanceof Civilian) {
				rescue((Civilian) targetHuman);
			}
			if (me().getPosition().equals(targetHuman.getPosition())) {
				if (!isInChangeSet(targetHuman.getID())) {
					targetHuman = null;
					return;
				}
				rescue(targetHuman.getID());
			}
			move(targetHuman.getPosition());
		}
	}


	protected void decide() throws ActionCommandException {
		// for bug of current zoning
		if (mySubZone != null && mySubZone.buildings.size() < 5) {
			mySubZone = zones[0].subZones[0];
		}
		checkEmptyBuildings();
		checkCivilianOnBoard();
		checkTarget();
		if (noCommi)
			noCommi();
		hearJob();
		checkHelp();
		selectCV();
		for (Entity entity : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian civilian = (Civilian) entity;
			if (model.getEntity(civilian.getPosition()) instanceof Building) {
				Building building = (Building) model.getEntity(civilian
						.getPosition());
				for (Building build : building.nearBuildings)
					if (build.lastTimeVisit == -1)
						build.lastTimeVisit = -2;
			}
		}
		if (reachableBuildings.size() >= modelBuildings.size() - 10)
			noBlockade();
		search();
	}

	private int targetCounter = 0;

	private void search() throws ActionCommandException {
		if (target != null
				&& (me().getPosition().equals(target) || !isReachable(target)))
			target = null;
		if (target == null) {
			int minTime = Integer.MAX_VALUE;
			for (Building building : reachableBuildings)
				if (building.isOnFire() == false
						&& building.lastTimeVisit < minTime) {
					minTime = building.lastTimeVisit;
					if (minTime == -2)
						break;
				}
			target = findTarget(minTime);
			if (target != null)
				targetCounter++;
		}

		if (target == null) {
			int minTime = Integer.MAX_VALUE;
			for (Area area : reachableAreas)
				if (area.lastTimeVisit < minTime) {
					minTime = area.lastTimeVisit;
					target = area.getID();
					if (minTime == -1)
						break;
				}
		}
		if (targetCounter >= 5 && reachableBuildings.size() < 10) {
			targetCounter = 0;
			int minTime = Integer.MAX_VALUE;
			for (rescuecore2.standard.entities.Road road : reachableRoads)
				if (road.lastTimeVisit < minTime) {
					minTime = road.lastTimeVisit;
					target = road.getID();
					if (minTime == -1)
						break;
				}
		}

		if (target == null) {
			randomWalk();
		} else
			move(target);
	}

	private EntityID findTarget(int minVisitTime) {
		wg.clearAreas();

		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		ArrayList<Enterance> start = wg.myEnterances;

		for (Enterance enterance : start)
			if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer) {
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.area.modelArea.lastTimeVisit <= minVisitTime
							&& (neighbour.area.modelArea instanceof Building || neighbour.area.modelArea instanceof Refuge)
							&& !((Building) neighbour.area.modelArea)
									.isOnFire())
						return neighbour.area.modelArea.getID();

					neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances) {
						if (!internal.mark) {
							if (internal.area.modelArea.lastTimeVisit <= minVisitTime
									&& (internal.area.modelArea instanceof Building || internal.area.modelArea instanceof Refuge)
									&& !((Building) internal.area.modelArea)
											.isOnFire())
								return internal.area.modelArea.getID();

							internal.mark = true;
							newLayer.add(internal);
						}
					}
				}
			}

			layer = newLayer;
		}

		return null;
	}

	private EntityID findTargetForNoBlockade(int minVisitTime) {
		wg.clearAreas();

		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		ArrayList<Enterance> start = wg.myEnterances;

		for (Enterance enterance : start)
			if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer) {
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.star
							&& neighbour.area.modelArea.lastTimeVisit <= minVisitTime
							&& (neighbour.area.modelArea instanceof Building || neighbour.area.modelArea instanceof Refuge)
							&& !((Building) neighbour.area.modelArea)
									.isOnFire())
						return neighbour.area.modelArea.getID();

					neighbour.mark = true;
					for (Enterance internal : neighbour.internalEnterances) {
						if (!internal.mark) {
							if (internal.star
									&& internal.area.modelArea.lastTimeVisit <= minVisitTime
									&& (internal.area.modelArea instanceof Building || internal.area.modelArea instanceof Refuge)
									&& !((Building) internal.area.modelArea)
											.isOnFire())
								return internal.area.modelArea.getID();

							internal.mark = true;
							newLayer.add(internal);
						}
					}
				}
			}

			layer = newLayer;
		}

		return null;
	}

	int count = 0;

	private void noBlockade() throws ActionCommandException {
		log("noblockade");
		if (mySubZone == null)
			return;
		log("1");
		int minTime = Integer.MAX_VALUE;
		while (count <= myZone1.subZones.length) {
			count++;
			for (Building building : modelBuildings)
				building.worldGraphArea.clearStar();
			for (Building building : mySubZone.buildings)
				building.worldGraphArea.setStar();
			for (Building building : mySubZone.buildings) {
				if (building.isOnFire() == false
						&& building.lastTimeVisit < minTime) {
					minTime = building.lastTimeVisit;
					if (minTime == -2)
						break;
				}
			}
			if (minTime > -1) {
				mySubZoneNumber = (mySubZoneNumber + 1)
						% myZone1.subZones.length;
				mySubZone = myZone1.subZones[mySubZoneNumber];
				continue;
			}
			break;
		}
		count = 0;
		EntityID target = findTargetForNoBlockade(minTime);
		if (target != null) {
			log("target: " + target.getValue());
			move(target);
		}
	}

	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.AMBULANCE_TEAM);
	}
}
