package agent;

import geometry.Degree;
import geometry.Line;
import geometry.Mathematic;
import geometry.Point;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Random;
import java.util.TreeMap;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import worldGraph.Blockade;
import worldGraph.Enterance;

public class FireBrigadeAgent extends Agent<FireBrigade> {
	private static final String MAX_DISTANCE_KEY = "fire.extinguish.max-distance";
	private static final String MAX_WATER_KEY = "fire.tank.maximum";
	private static final String MAX_POWER_KEY = "fire.extinguish.max-sum";
	private int maxWater = -1;
	private int maxDistance = -1;
	private int maxPower = -1;
	private Building extTarget;
	private FireZone myZone = null;

	public Point centralPointOfMap = new Point();

	protected void postConnect() {
		super.postConnect();
		maxWater = config.getIntValue(MAX_WATER_KEY);
		maxDistance = config.getIntValue(MAX_DISTANCE_KEY);
		maxPower = config.getIntValue(MAX_POWER_KEY);
		long m = System.nanoTime();
		System.out.println("PreCompuation of Fire Started");
		int x1 = 0, y1 = 0;
		for (Building building : modelBuildings) {
			x1 += building.getX();
			y1 += building.getY();
		}
		centralPointOfMap = new Point(x1 / modelBuildings.size(), y1
				/ modelBuildings.size());

		for (Building building : modelBuildings)
			for (Road road : building.nearRoads) {
				ArrayList<Point> visionPoints = getVisionPoints(road, building);
				if (visionPoints != null && visionPoints.size() != 0)
					building.visionRoads.put(road, visionPoints);
			}
		for (Building building : modelBuildings)
			for (EntityID e : building.getNeighbours()) {
				Entity entity = model.getEntity(e);
				if (entity instanceof Building) {
					building.visionBuildings.add((Building) entity);
				}
			}

//		Building building = (Building) model.getEntityByInt(54436);
//		Road road = (Road) model.getEntityByInt(35368);
//		ArrayList<Point> points = getVisionPoints(road, building);
//		for (Point point : points)
//		{
//			System.out.println("circle(" + point.getX() + ", " + point.getY() + ") 500, 4;");
//		}
//		System.exit(0);
		long l = System.nanoTime();
		long d = (l - m) / 1000000;
		System.out.println(d);
		System.out.println("PreCompuation of Fire Finished");
	}

	public ArrayList<Point> getVisionPoints(Area area, Building building) {
		return getVisionPoints(model, area, building, maxDistance);
	}

	public static ArrayList<Point> getVisionPoints(StandardWorldModel model,
			Area area, Building building, int maxDistance) {
		maxDistance *= 4;
		float minDegree = Integer.MAX_VALUE, maxDegree = -Integer.MAX_VALUE;
		ArrayList<Point> roadPoints = area.worldGraphArea.points;
		ArrayList<Point> visionPoints = new ArrayList<Point>();

		Point buildingPoint = new Point(building.getX(), building.getY());
		Point minPoint = null, maxPoint = null;
		for (int i = 0; i < roadPoints.size() - 1; i++) {
			Line line = new Line(buildingPoint, roadPoints.get(i));
			float degree = Degree.absoluteAngle(line.getTheta());

			if (i == 0) {
				minDegree = maxDegree = degree;
				minPoint = maxPoint = roadPoints.get(i);
			} else {
				if (Degree.isClockWise(roadPoints.get(i - 1), buildingPoint,
						roadPoints.get(i))) {
					if (Degree.isClockWise(minPoint, buildingPoint,
							roadPoints.get(i))) {
						minDegree = degree;
						minPoint = roadPoints.get(i);
					}
				} else {
					if (!Degree.isClockWise(maxPoint, buildingPoint,
							roadPoints.get(i))) {
						maxDegree = degree;
						maxPoint = roadPoints.get(i);
					}
				}
			}
		}
		int m = 3;
		if (area.worldGraphArea.points.size() < 7)
			m = 1;

		ArrayList<Building> nearBuildings = new ArrayList<Building>();
		for (Area a : building.nearAreas50000)
			if (a instanceof Building)
				nearBuildings.add((Building) a);
		float deg1 = Degree.absoluteAngle((minDegree + maxDegree) / 2);
		for (int j = 0; j < m; j++) {
			if (j == 1)
				deg1 = Degree.absoluteAngle((minDegree + maxDegree) / 2
						- (Degree.getDeltaAngle(maxDegree, minDegree)) / 4);
			if (j == 2)
				deg1 = Degree.absoluteAngle((minDegree + maxDegree) / 2
						+ (Degree.getDeltaAngle(maxDegree, minDegree)) / 4);
			float deg2 = Degree.absoluteAngle((float) Math.PI + deg1);
			float degree = deg1;
			Point tmpPoint = new Point(building.getX() + 10
					* (int) Math.cos(deg2), building.getY() + 10
					* (int) Math.sin(deg2));
			if (Degree.isClockWise(tmpPoint, buildingPoint, minPoint)
					&& !Degree.isClockWise(tmpPoint, buildingPoint, maxPoint))
				degree = deg2;

			int x = (int) (building.getX() + maxDistance * Math.cos(degree));
			int y = (int) (building.getY() + maxDistance * Math.sin(degree));
			Line buildingToRoad = new Line(buildingPoint, new Point(x, y));
			float minDist = Integer.MAX_VALUE, maxDist = -Integer.MAX_VALUE;
			for (int i = 0; i < roadPoints.size() - 1; i++) {
				Line line = new Line(roadPoints.get(i), roadPoints.get(i + 1));
				Point intersectPoint = buildingToRoad.getIntersectPoint(line);
				if (intersectPoint != null) {
					float dist = intersectPoint.getDistance(buildingPoint);
					if (dist < minDist)
						minDist = dist;
					if (dist > maxDist)
						maxDist = dist;
				}
			}
			if (maxDist < 0)
				continue;
			if (minDist > MAX_DISTANCE_VIEW)
				continue;
			maxDist = Math.min(maxDist, MAX_DISTANCE_VIEW);
			for (int i = 0; i < 3; i++) {
				float dist = (minDist + maxDist) / 2;
				if (i == 1)
					dist = maxDist - 500; // TAVAJJOH
				if (i == 2)
					dist = minDist + 500; // TAVAJJOH
				int newX = (int) (building.getX() + dist * Math.cos(degree));
				int newY = (int) (building.getY() + dist * Math.sin(degree));
				Point newPoint = new Point(newX, newY);
				buildingToRoad.setSecondPoint(newPoint);

				boolean hasIntersect = false;
				done: for (Building bld : nearBuildings)
					for (Edge edge : bld.getEdges())
						if (edge.getNeighbour() == null) {
							Line lineEdge = new Line(new Point(
									edge.getStartX(), edge.getStartY()),
									new Point(edge.getEndX(), edge.getEndY()));
							if (Line.isIntersectBetweenLines(buildingToRoad,
									lineEdge)) {
								hasIntersect = true;
								break done;
							}
						}

				if (!hasIntersect)
					visionPoints.add(buildingToRoad.getSecondPoint());
			}
		}

		return visionPoints;
	}

	public EntityID target = null;

	private void checkWater() throws ActionCommandException {
		double minDist = Integer.MAX_VALUE;
		Refuge minRefuge = null;
		int minlts = Integer.MAX_VALUE;
		if (modelRefuges.size() == 0 && me().getWater() == 0)
			search();
		else {
			if (model.getEntity(me().getPosition()) instanceof Refuge
					&& me().getWater() != maxWater) {
				extTarget = null;
				rest();
			}

			if (me().getWater() == 0) {
				Refuge refuge = null;
				if (reachableRefuges != null && reachableRefuges.size() != 0)
					for (Refuge refuge2 : reachableRefuges) {
						if (Math.hypot(me().getX() - refuge2.getX(), me()
								.getY() - refuge2.getY()) <= minDist) {
							minDist = Math.hypot(me().getX() - refuge2.getX(),
									me().getY() - refuge2.getY());
							minRefuge = refuge2;
						}
					}
				else {
					for (Refuge refuge2 : modelRefuges) {
						if (refuge2.lastTimeSeen < minlts) {
							minlts = refuge2.lastTimeSeen;
							minRefuge = refuge2;
						}
					}
					move(minRefuge.getID(), false);
				}

				refuge = minRefuge;
				if (refuge != null)
					move(refuge.getID());
				search();
			}
		}
	}

	public static final double w2 = 1, w1 = 0, w3 = -10;
	// w2 = 0.01, w1 = 0.8, w3 = -100
	private ArrayList<Building> reachableAndIgnitedBuildings = new ArrayList<Building>();

	private int counter = 10;

	private Building selectTarget() {
		if (myZone == null)
			return null;
		counter++;
		if (extTarget == null
				|| !buildingReachability(extTarget)
				|| !extTarget.isOnFire()
				|| (!commandHistory.get(time - 1).equals(
						StandardMessageURN.AK_EXTINGUISH) && counter >= 5)
				|| counter >= 10) {
			counter = 0;
			if (model.getEntity(me().getPosition()) instanceof Building
					&& ((Building) model.getEntity(me().getPosition()))
							.isOnFire()) {
				reachableAndIgnitedBuildings.add((Building) model
						.getEntity(me().getPosition()));
			}

			double minevalue = Double.MAX_VALUE;
			Building minBuil = null;
			ArrayList<Building> aroundReachable = new ArrayList<Building>();
			for (Building building : myZone.aroundBuildings)
				if (buildingReachability(building))
					aroundReachable.add(building);
			if (aroundReachable.size() != 0) {
				for (Building buil : aroundReachable)
					if ((minevalue == Double.MAX_VALUE || buil.value < minevalue)) {
						minevalue = buil.value;
						minBuil = buil;
					}
			} else {
				FireZone tmpZone = new FireZone(myZone.zoneNum);
				for (Building building : myZone.ignitedBuilds)
					if (buildingReachability(building))
						tmpZone.ignitedBuilds.add(building);
				tmpZone.time = time;
				tmpZone.logger = logger;
				tmpZone.setAroundBuildings();
				for (Building building : tmpZone.aroundBuildings)
					setValueForBuildings(building);
				if (tmpZone.aroundBuildings.size() != 0)
					for (Building building : tmpZone.aroundBuildings) {

						if ((minevalue == Double.MAX_VALUE || building.value < minevalue)
								&& buildingReachability(building)) {
							minevalue = building.value;
							minBuil = building;
						}
					}
				else
					minBuil = null;
			}

			return minBuil;
		}
		return extTarget;
	}

	private boolean buildingReachability(Building building) {
		if (building.worldGraphArea.buildingReachabilityCalculated)
			return building.worldGraphArea.buildingReachability;
		building.worldGraphArea.buildingReachabilityCalculated = true;
		if (isInChangeSet(building.getID())) {
			building.worldGraphArea.buildingReachability = true;
			return true;
		}
		if (building.visionRoads.size() == 0
				&& building.visionBuildings.size() == 0) {
			building.worldGraphArea.buildingReachability = false;
			return false;
		}
		if (building.worldGraphArea.isReachable) {
			building.worldGraphArea.buildingReachability = true;
			return true;
		}
		for (Road road : building.visionRoads.keySet())
			if (road.worldGraphArea.isReachable) {
				for (Point point : building.visionRoads.get(road)) {
					boolean i = true;
					for (Blockade blockade : road.worldGraphArea.blockades)
						if (blockade.isInShape(point)) {
							i = false;
							break;
						}
					if (i == true) {
						building.worldGraphArea.buildingReachability = true;
						return true;
					}
				}
			}
		for (Building b : building.visionBuildings)
			if (b.worldGraphArea.isReachable) {
				building.worldGraphArea.buildingReachability = true;
				return true;
			}

		building.worldGraphArea.buildingReachability = false;
		return false;
	}

	private void moveAndExtinguish(Building fireTarget)
			throws ActionCommandException {
		if (fireTarget != null) {
			if (isInChangeSet(fireTarget.getID()) && canExtinguish(fireTarget))
				extinguish(fireTarget.getID(), maxPower);

			moveToFire(fireTarget);
		}
	}

	private GroupForFire getMinPathForFireBrigade(ArrayList<Enterance> start) {
		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		wg.clearAreas();
		GroupForFire groupForFires = new GroupForFire();
		for (Enterance enterance : start)
			if (enterance.area.target4Fire
					&& enterance.area.checkSituation4Fire == false) {
				if (enterance.area.fire.size() > 0) {
					for (int i = 0; i < enterance.area.fire.size(); i++) {
						groupForFires.fireAgents
								.add(enterance.area.fire.get(i));
						enterance.area.fire.get(i).fireTrueOrFalse = true;
					}
				}
				if (enterance.area.zoneNum != -1)
					groupForFires.zoneNum.add(enterance.area.zoneNum);

				enterance.area.checkSituation4Fire = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer)
				if (enterance.isItConnectedToNeighbour
						&& !enterance.neighbour.area.checkSituation4Fire) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.area.target4Fire) {
						if (neighbour.area.fire.size() > 0) {
							for (int i = 0; i < neighbour.area.fire.size(); i++) {
								groupForFires.fireAgents
										.add(neighbour.area.fire.get(i));
								neighbour.area.fire.get(i).fireTrueOrFalse = true;
							}
						}
						if (neighbour.area.zoneNum != -1)
							groupForFires.zoneNum.add(neighbour.area.zoneNum);
					}

					neighbour.area.checkSituation4Fire = true;
					for (Enterance internal : neighbour.internalEnterances)
						if (!internal.area.checkSituation4Fire) {
							internal.lastEnterance = neighbour;
							if (internal.area.target4Fire) {
								if (internal.area.fire.size() > 0) {
									for (int i = 0; i < internal.area.fire
											.size(); i++) {
										groupForFires.fireAgents
												.add(internal.area.fire.get(i));
										internal.area.fire.get(i).fireTrueOrFalse = true;
									}
								}
								if (internal.area.zoneNum != -1)
									groupForFires.zoneNum
											.add(internal.area.zoneNum);
							}

							internal.area.checkSituation4Fire = true;
							newLayer.add(internal);
						}
				}

			layer = newLayer;
		}

		return groupForFires;
	}

	private void setWorkersNum() {

		for (FireZone fireZone : modelZones) {
			if (fireZone.aroundBuildings.size() <= 5) {
				fireZone.workersNum = 3;
				break;
			}
			if (fireZone.aroundBuildings.size() <= 10) {
				fireZone.workersNum = 6;
				break;
			}
			if (fireZone.aroundBuildings.size() <= 15) {
				fireZone.workersNum = 9;
				break;
			}
			if (fireZone.aroundBuildings.size() > 15) {
				fireZone.workersNum = 15;
				break;
			}
		}

	}

	private void setFireIdAndZoneNum() {
		for (Area area : modelAreas) {
			area.worldGraphArea.zoneNum = 0;
			area.worldGraphArea.fire = new ArrayList<FireBrigade>();
		}
		// TODO optimize
		for (Area area : modelAreas) {
			for (FireBrigade fireBrigade : modelFireBrigades) {
				if (area.getID().getValue() == fireBrigade.getPosition()
						.getValue() && fireBrigade.hasBuriedness == false) {
					area.worldGraphArea.fire.add(fireBrigade);
					area.worldGraphArea.target4Fire = true;
				}
			}
			if (area instanceof Building) {
				if (((Building) area).zoneNumber >= 0) {
					area.worldGraphArea.zoneNum = ((Building) area).zoneNumber;
					area.worldGraphArea.target4Fire = true;
				}
			}
		}
	}

	public TreeMap<ArrayList<Integer>, ArrayList<Integer>> treeMap = new TreeMap<ArrayList<Integer>, ArrayList<Integer>>();
	public TreeMap<FireBrigade, Building> FB = new TreeMap<FireBrigade, Building>();
	public GroupForFire groupForFire = new GroupForFire();
	public int me = 0;
	public int t = 100, d = 50;

	public int countReachableBuilds = 0;

	protected void decide() throws ActionCommandException {
		countReachableBuilds = 0;

		for (Building building : modelBuildings)
			if (buildingReachability(building))
				countReachableBuilds++;

		for (EntityID e : changeSet.getChangedEntities())
			if (model.getEntity(e) instanceof Building) {
				Building building = (Building) model.getEntity(e);
				if (building.isOnFire())
					for (Building b : building.nearBuildings)
						if (b.lastTimeSeen < 0
								&& (!b.isFierynessDefined() || b.getFieryness() == 0))
							b.lastTimeSeen = LTSDORI;
			}

		initialBuildings();
		for (Area ar : modelAreas)
			ar.worldGraphArea.fB.clear();

		for (FireBrigade fire : modelFireBrigades) {
			worldGraph.Area area = (((Area) model.getEntityByInt(fire
					.getPosition().getValue())).worldGraphArea);
			area.fB.add(fire);
		}
		zoning();
		for (FireZone fireZone : modelZones) {
			fireZone.time = time;
			fireZone.logger = logger;
			fireZone.setAroundBuildings();
		}
		setValueForAroundBuildings();
		if (modelZones.size() != 0)
			for (FireZone fireZone : modelZones) {
				int allX = 0, allY = 0;
				for (Building building : fireZone.buildings) {
					allX += building.getX();
					allY += building.getY();
				}
				fireZone.centerX = allX / fireZone.buildings.size();
				fireZone.centerY = allY / fireZone.buildings.size();
			}
		setWorkersNum();
		setFireIdAndZoneNum();
		for (FireBrigade fireBrigade : modelFireBrigades)
			fireBrigade.fireTrueOrFalse = false;

		if (me().fireTrueOrFalse == false) {
			worldGraph.Area fireArea = (((Area) model.getEntityByInt(me()
					.getPosition().getValue())).worldGraphArea);
			groupForFire = getMinPathForFireBrigade(fireArea.enterances);
			me().fireTrueOrFalse = true;
		}

		for (FireZone fireZone : modelZones) {
			if (fireZone.ignitedBuilds.size() <= Integer.parseInt(psdConfig
					.get("Fire", "Zone", "InhibiterBuildingsCount")))
				fireZone.isInhibitor = true;
		}
		if (noCommi)
			selectMyZoneForNoCommi();
		else
			selectMyZone();
		checkWater();
		checkEmergencySituation();
		Building fireTarget = selectTarget();
		if (extTarget != null && !extTarget.isOnFire())
			extTarget = null;
		if (fireTarget != null)
			extTarget = fireTarget;
		moveAndExtinguish(fireTarget);
		unemploymentSituation();
		if (countReachableBuilds > (modelBuildings.size() - 10))
			noBlockadeSearch();
		search();
	}

		private HashSet<FireBrigade> bFSForFire() {
		HashSet<Enterance> laye = new HashSet<Enterance>();
		HashSet<FireBrigade> list = new HashSet<FireBrigade>();
		wg.clearAreas();
		list.add(me());
		for (Enterance enterance : wg.myEnterances) {
			if (!enterance.area.checkSituation4Fire
					&& enterance.area.fB != null
					&& enterance.area.fB.size() > 0) {
				enterance.area.checkSituation4Fire = true;
				list.addAll(enterance.area.fB);
			}

			laye.add(enterance);
		}
		while (laye.size() > 0) {
			HashSet<Enterance> newLaye = new HashSet<Enterance>();
			for (Enterance enterance : laye) {
				Enterance neighbourE = enterance.neighbour;
				if (!neighbourE.area.checkSituation4Fire
						&& enterance.isItConnectedToNeighbour) {
					if (neighbourE.area.fB != null
							&& neighbourE.area.fB.size() != 0)
						list.addAll(neighbourE.area.fB);
				}
				enterance.mark = true;
				neighbourE.area.checkSituation4Fire = true;
				neighbourE.mark = true;
				for (Enterance internal : neighbourE.internalEnterances)
					if (!internal.mark)
						newLaye.add(internal);
			}

			laye = newLaye;
		}
		return list;
	}

	private void selectMyZoneForNoCommi() {
		ArrayList<FireZone> zones = new ArrayList<FireZone>();
		FireZone fireZone1 = null;
		int minSize = Integer.MAX_VALUE;
		if (me().hasBuriedness)
			return;
		myZone = null;
		if (modelZones.size() != 0)
			for (FireZone fireZone : modelZones) {
				zones.add(fireZone);
			}
		else
			return;

		for (FireZone fireZone : zones) {
			if (fireZone.ignitedBuilds.size() < minSize) {
				minSize = fireZone.ignitedBuilds.size();
				fireZone1 = fireZone;
			}
		}
		myZone = fireZone1;
	}

	public Comparator<FireBrigade> idComparator = new Comparator<FireBrigade>() {
		public int compare(FireBrigade a, FireBrigade b) {
			if (a.getID().getValue() > b.getID().getValue())
				return 1;
			if (a.getID().getValue() < b.getID().getValue())
				return -1;
			return 0;
		}
	};
	public Comparator<FireZone> zoneComparator = new Comparator<FireZone>() {
		public int compare(FireZone a, FireZone b) {
			if (a.value > b.value)
				return 1;
			if (a.value < b.value)
				return -1;
			return 0;
		}
	};

	private void selectMyZone() {
		ArrayList<FireBrigade> fireBrigades = new ArrayList<FireBrigade>();
		ArrayList<FireZone> hibitorFireZones = new ArrayList<FireZone>();
		ArrayList<FireZone> otherZones = new ArrayList<FireZone>();
		for (FireBrigade brigade : bFSForFire()) {
			if (!brigade.hasBuriedness)
				fireBrigades.add(brigade);
		}
		if (me().hasBuriedness)
			return;
		myZone = null;
		if (map == MapType.Berlin || map == MapType.Paris || map==MapType.Istanbul) {
			FireZone smallestZone = null;
			int minvalue = Integer.MAX_VALUE;
			for (FireZone fireZone : modelZones)
				fireZone.volume += fireZone.buildings.size();
			for (FireZone fireZone : modelZones)
				fireZone.value = 10000
						* (fireZone.buildings.size())
						+ (int) Math.hypot(me().getX() - fireZone.centerX, me()
								.getY() - fireZone.centerY) + fireZone.volume;
			for (FireZone fireZone : modelZones)
				if (fireZone.value < minvalue) {
					smallestZone = fireZone;
					minvalue = fireZone.value;
				}
			myZone = smallestZone;
		} else {
			for (FireZone fireZone : modelZones) {
				if (fireZone.isInhibitor)
					hibitorFireZones.add(fireZone);
				else
					otherZones.add(fireZone);
			}
			int w = 0;

			Collections.sort(fireBrigades, idComparator);
			if (hibitorFireZones.size() != 0)
				for (FireZone zone : hibitorFireZones) {
					for (int i = w; i < zone.workersNum + 2; i++) {
						if (me().getID().getValue() == fireBrigades.get(i)
								.getID().getValue()) {
							myZone = zone;
							return;
						}
					}
					w += zone.workersNum + 2;
				}

			for (FireZone fireZone1 : otherZones)
				fireZone1.value = (int) (2 * fireZone1.aroundBuildings.size() + Math
						.hypot(me().getX() - fireZone1.centerX, me().getY()
								- fireZone1.centerY));

			Collections.sort(otherZones, zoneComparator);

			if (otherZones.size() != 0) {
				for (FireZone fZone : otherZones)
					for (int i = w; i < zone.workersNum + 2; i++) {
						if (me().getID().getValue() == fireBrigades.get(i)
								.getID().getValue()) {
							myZone = fZone;
							return;
						}
					}
				w += zone.workersNum + 2;
			}
			FireZone minZone = null;
			double minDist = Integer.MAX_VALUE, dist;
			for (FireZone z : modelZones) {
				dist = Math.hypot(me().getX() - z.centerX, me().getY()
						- z.centerY);
				if (dist <= minDist) {
					minDist = dist;
					minZone = z;
				}
			}

			myZone = minZone;
		}
	}

	public int getDistToFire(Building building) {
		wg.clearTargets();
		int minDist = Integer.MAX_VALUE;
		for (Road road : building.visionRoads.keySet()) {
			if (road.worldGraphArea.isReachable)
				for (Point point : building.visionRoads.get(road)) {
					boolean i = true;
					for (Blockade blockade : road.worldGraphArea.blockades)
						if (blockade.isInShape(point)) {
							i = false;
							break;
						}
					if (i == true) {
						if (me().getPosition().getValue() == road.getID()
								.getValue())
							return 1;
						if (road.worldGraphArea.distanceFromSelf < minDist)
							minDist = road.worldGraphArea.distanceFromSelf;
						break;
					}
				}
		}
		for (Building buil : building.visionBuildings) {
			if (buil.worldGraphArea.isReachable)
				if (me().getPosition().getValue() == buil.getID().getValue())
					return 1;
			if (buil.worldGraphArea.distanceFromSelf < minDist)
				minDist = buil.worldGraphArea.distanceFromSelf;
			break;
		}

		if (minDist == Integer.MAX_VALUE)
			return 1000;
		return minDist;
	}

	private void moveToFire(Building target) throws ActionCommandException {
		wg.clearTargets();
		log("Target: " + target);
		for (Road road : target.visionRoads.keySet()) {
			if (road.worldGraphArea.isReachable)
				for (Point point : target.visionRoads.get(road)) {
					boolean i = true;
					for (Blockade blockade : road.worldGraphArea.blockades)
						if (blockade.isInShape(point)) {
							i = false;
							break;
						}
					if (i == true) {
						if (me().getPosition().getValue() == road.getID()
								.getValue()) {
							ArrayList<EntityID> path = new ArrayList<EntityID>();
							path.add(road.getID());
							move(path, point.getX(), point.getY());
						}
						road.worldGraphArea.setAreaAsTarget(true);
						break;
					}
				}
		}
		for (Building building : target.visionBuildings) {
			if (building.worldGraphArea.isReachable)
				if (me().getPosition().getValue() == building.getID()
						.getValue()) {
					ArrayList<EntityID> path = new ArrayList<EntityID>();
					path.add(building.getID());
					move(path, building.getX(), building.getY());
				}
			building.worldGraphArea.setAreaAsTarget(true);
			break;
		}

		int x = -1, y = -1;
		ArrayList<EntityID> path = wg.getMinPath(wg.myEnterances, true);
		if (path.size() == 0 && target.worldGraphArea.isReachable)
			move(target.getID());
		for (Road r : target.visionRoads.keySet())
			if (path.get(path.size() - 1).getValue() == r.getID().getValue()) {
				for (Point point : target.visionRoads.get(r)) {
					boolean i = true;
					for (Blockade blockade : r.worldGraphArea.blockades)
						if (blockade.isInShape(point)) {
							i = false;
							break;
						}
					if (i == true) {
						x = point.getX();
						y = point.getY();
						move(path, x, y);
					}
				}
			}
		for (Building b : target.visionBuildings)
			if (path.get(path.size() - 1).getValue() == b.getID().getValue()) {
				x = b.getX();
				y = b.getY();
				move(path, x, y);
			}
	}

	private boolean canExtinguish(Building target) {
		return Mathematic.getDistance(target.getX(), target.getY(),
				me().getX(), me().getY()) < maxDistance;
	}

	private int akhir = 0, targetCounter = 0;
	private int[] position = new int[3];
	private Building tar = null;

	private void search() throws ActionCommandException {
		position[akhir] = me().getPosition().getValue();
		akhir++;
		if (akhir == 3)
			akhir = 0;
		if (position[0] == position[1] && position[1] == position[2]
				&& reachableRoads.size() > 0) {
			int i = (new Random()).nextInt(reachableRoads.size());
			target = reachableRoads.get(i).getID();
			position[0] = 1;
			position[1] = 2;
			position[2] = 3;
		}

		if (reachableBuildings.size() > 10) {
			int minlts1 = Integer.MAX_VALUE;

			ArrayList<Building> reachableBuildVisionRoads = new ArrayList<Building>();
			Building tar2 = null;
			for (Building building : modelBuildings)
				if (buildingReachability(building))
					reachableBuildVisionRoads.add(building);

			for (Building building : reachableBuildVisionRoads)
				if (building.lastTimeSeen <= minlts1) {
					minlts1 = building.lastTimeSeen;
					tar2 = building;
				}

			if (tar == null || !buildingReachability(tar)
					|| (tar2 != null && tar2.lastTimeSeen < tar.lastTimeSeen))
				tar = tar2;

			if (tar != null)
				moveToFire(tar);
		}

		if (target != null
				&& (me().getPosition().equals(target) || !((Area) (model
						.getEntity(target))).worldGraphArea.isReachable))
			target = null;

		if (target == null) {
			int minlts = Integer.MAX_VALUE;
			for (Building b : reachableBuildings) {
				if (b.lastTimeSeen < minlts) {
					minlts = b.lastTimeSeen;
					if (minlts == LTSDORI)
						break;
				}
			}

			target = findTarget(minlts);
			if (target != null)
				targetCounter++;
		}
		if (target == null) {
			int minltsa = Integer.MAX_VALUE;
			for (Area a : reachableAreas)
				if (a.lastTimeSeen < minltsa) {
					minltsa = a.lastTimeSeen;
					target = a.getID();
					if (minltsa == LTSDORI)
						break;
				}
		}
		if (targetCounter == 5) {
			targetCounter = 0;
			int minTime = Integer.MAX_VALUE;
			for (rescuecore2.standard.entities.Road road : reachableRoads)
				if (road.lastTimeSeen < minTime) {
					minTime = road.lastTimeSeen;
					target = road.getID();
					if (minTime == LTSDORI)
						break;
				}
		}

		if (target == null) {
			shouldBaseDecideForRandomWalk = false;
			randomWalk();
		} else
			move(target);
	}

	private void noBlockadeSearch() throws ActionCommandException {
		if (myZone1 == null)
			return;
		position[akhir] = me().getPosition().getValue();
		akhir++;
		if (akhir == 3)
			akhir = 0;
		if (position[0] == position[1] && position[1] == position[2]) {
			if (reachableRoads.size() > 0) {
				int i = (new Random()).nextInt(reachableRoads.size());
				target = reachableRoads.get(i).getID();
				position[0] = 1;
				position[1] = 2;
				position[2] = 3;
			}
		}

		Building tar2 = null;
		while (true) {
			int minlts1 = Integer.MAX_VALUE;
			ArrayList<Building> reachableBuildHa = new ArrayList<Building>();

			for (Building building : myZone1.buildings)
				if (buildingReachability(building))
					reachableBuildHa.add(building);

			for (Building building : reachableBuildHa)
				if (building.lastTimeSeen <= minlts1) {
					minlts1 = building.lastTimeSeen;
					tar2 = building;
				}

			if (minlts1 > -1) {
				myZoneNumber = (myZoneNumber + 1) % zones.length;
				myZone1 = zones[myZoneNumber];
				for (Building building : myZone1.buildings)
					building.lastTimeSeen = -1;
				continue;
			}
			break;
		}

		if (tar == null || !tar.worldGraphArea.isReachable
				|| (tar2 != null && tar2.lastTimeSeen < tar.lastTimeSeen))
			tar = tar2;

		if (tar != null)
			moveToFire(tar);
	}

	private EntityID findTarget(int minlst1) {
		wg.clearAreas();
		ArrayList<Enterance> start = wg.myEnterances;
		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		for (Enterance myenter : start) {
			if (!myenter.mark) {
				myenter.mark = true;
				layer.add(myenter);
			}

			while (layer.size() > 0) {
				ArrayList<Enterance> layer1 = new ArrayList<Enterance>();
				for (Enterance myenter1 : layer)
					if (myenter1.isItConnectedToNeighbour
							&& !myenter1.neighbour.mark) {
						Enterance neighbour = myenter1.neighbour;

						if (neighbour.area.modelArea.lastTimeSeen <= minlst1
								&& neighbour.area.modelArea instanceof Building)
							return myenter1.area.modelArea.getID();
						neighbour.mark = true;

						for (Enterance internal : neighbour.internalEnterances) {
							if (!internal.mark) {
								if (internal.area.modelArea.lastTimeSeen <= minlst1
										&& internal.area.modelArea instanceof Building)
									return neighbour.area.modelArea.getID();
								internal.mark = true;
								layer1.add(internal);
							}
						}
					}
				layer = layer1;
			}
			return null;
		}
		return target;
	}

	private int valueForNotIgnited(int v) {
		int[] arr = { 0, 1, 5, 15, 30 };
		for (int i = 0; i < arr.length; i++)
			if (v <= arr[i])
				return i;

		return arr.length;
	}

	private void setValueForBuildings(Building build) {
		build.value = w1
				* build.getGroundArea()
				* build.getFloors()
				+ w2
				* getDistToFire(build)
				+ w3
				* (valueForNotIgnited(build.notIgnitedNearBuildings.size()) + 1);
	}

	private void setValueForAroundBuildings() {
		for (FireZone fireZone : modelZones)
			for (Building build : fireZone.aroundBuildings)
				if (buildingReachability(build))
					setValueForBuildings(build);
	}

	private void initialBuildings() {
		ArrayList<Building> ignitedBuildings = new ArrayList<Building>();
		reachableAndIgnitedBuildings = new ArrayList<Building>();
		for (Building build : modelBuildings) {
			if (build.isOnFire()) {
				build.notIgnitedNearBuildings = new ArrayList<Building>();
				ignitedBuildings.add(build);
				for (Building buil : build.nearBuildings)
					if (!buil.isFierynessDefined() || buil.getFieryness() == 0) {
						build.notIgnitedNearBuildings.add(buil);
					}
			}
		}

		for (Building build : ignitedBuildings)
			if (buildingReachability(build))
				reachableAndIgnitedBuildings.add(build);
	}

	private Building tar1 = null;

	private void unemploymentSituation() throws ActionCommandException {
		Building buil1 = null;
		int maxLast = 0;
		for (StandardEntity e : model.getObjectsInRange(me().getX(), me()
				.getY(), MAX_DISTANCE_EXTINGUISH))
			if (e instanceof Building && ((Building) e).isOnFire()
					&& !buildingReachability((Building) e)
					&& ((Building) e).lastTimeSeen > maxLast
					&& ((Building) e).lastTimeSeen + 5 >= time) {
				maxLast = ((Building) e).lastTimeSeen;
				buil1 = ((Building) e);
			}
		if (buil1 != null)
			extinguish(buil1.getID(), maxPower);
		ArrayList<Building> unreachableFiredBuildings = new ArrayList<Building>();
		for (Building building : modelBuildings)
			if (building.isOnFire() && !buildingReachability(building)
					&& building.lastTimeSeen + 10 >= time)
				for (Area area : building.nearAreas50000)
					if (area.worldGraphArea.isReachable) {
						unreachableFiredBuildings.add(building);
						break;
					}
		int maxheard = 0;
		Building tar = null;
		for (Building building : unreachableFiredBuildings)
			if (building.lastTimeSeen > maxheard) {
				maxheard = building.lastTimeSeen;
				tar = building;
			}
		if (tar1 == null || !tar1.isOnFire()
				|| !unreachableFiredBuildings.contains(tar1))
			tar1 = tar;
		if (tar1 != null) {
			wg.clearTargets();
			for (Area area : tar1.nearAreas50000)
				if (area instanceof Road)
					area.worldGraphArea.setAreaAsTarget(true);
			ArrayList<EntityID> path = wg.getMinPath(wg.myEnterances, true);
			move(path);
		}
	}

	private void checkEmergencySituation() throws ActionCommandException {
		if (reachableAreas.size() <= 6)
			for (Building build : modelBuildings)
				if (build.isOnFire()
						&& Math.hypot(me().getX() - build.getX(), me().getY()
								- build.getY()) <= maxDistance)
					extinguish(build.getID(), maxPower);
	}

	public ArrayList<FireZone> modelZones = null;
	private FireZone zone = null;

	public void zoning() {
		modelZones = new ArrayList<FireZone>();
		for (Building isCheckingbuilding : modelBuildings) {
			isCheckingbuilding.zoneNumber = -1;
			isCheckingbuilding.haveZone = false;
		}

		for (Building isCheckingbuilding : reachableAndIgnitedBuildings) {
			if (!isCheckingbuilding.haveZone) {
				zone = new FireZone(modelZones.size());
				checkingNeighboursForZoning(isCheckingbuilding);
				modelZones.add(zone);
			}
		}
	}

	private void checkingNeighboursForZoning(Building isChecking) {
		isChecking.zoneNumber = zone.zoneNum;
		zone.buildings.add(isChecking);
		if (isChecking.isOnFire())
			zone.ignitedBuilds.add(isChecking);
		isChecking.haveZone = true;
		for (Building neighbour : isChecking.nearBuildings)
			if (((neighbour.isFierynessDefined() && neighbour.getFieryness() != 0) || neighbour
					.isOnFire()) && !neighbour.haveZone)
				checkingNeighboursForZoning(neighbour);
	}

	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}
}
