package agent;

import geometry.Line;
import geometry.Mathematic;
import geometry.Point;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import worldGraph.Enterance;

public class PoliceForceAgent extends Agent<PoliceForce> {
	private static final String DISTANCE_KEY = "clear.repair.distance";
	private int distance;
	private ArrayList<Integer> clearedIDs = new ArrayList<Integer>();
	public static int MAX_DIST_FOR_HELP = 30000;

	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.POLICE_FORCE);

	}

	protected void postConnect() {
		super.postConnect();
		distance = config.getIntValue(DISTANCE_KEY);
		for (Building building : modelBuildings)
			for (Road road : building.nearRoads) {
				ArrayList<Point> visionPoints = FireBrigadeAgent
						.getVisionPoints(model, road, building, MAX_DISTANCE_VIEW);
				if (visionPoints != null && visionPoints.size() != 0)
					building.visionRoads.put(road, visionPoints);
			}
		for (Building building : modelBuildings)
			for (EntityID e : building.getNeighbours()) {
				Entity entity = model.getEntity(e);
				if (entity instanceof Building)
					building.visionBuildings.add((Building) entity);
			}
	}

	int myTarget = -1;
	int heardTarget = -1;
	int FIRST_ATTACK = 70;
	int SECOND_ATTACK = 200;

	enum SearchMode {
		SEARCH_IN_POLICEZONE, SEARCH_IN_SUBZONE
	}

	boolean haveReachableRefuge = false;

	protected void checkHelpIn(SearchMode searchMode)
			throws ActionCommandException {
		if (me().hasBuriedness || mySubZone == null)
			return;

		log("start check help in....!!!!!+D+D+D+D+D+D++D+DD+");
		// After First Attack
		ArrayList<Integer> targs = new ArrayList<Integer>();
		for (int nhP : longnHP.values()) {
			Entity nhpE = model.getEntityByInt(nhP);
			if (!clearedIDs.contains(nhP)) {
				if (changeSet.getChangedEntities().contains(nhpE.getID())
						&& ((Area) nhpE).worldGraphArea.blockades.size() == 0
						&& isReachable(nhpE.getID())) {
					clearedIDs.add(nhP);
					continue;
				}

				if (searchMode == SearchMode.SEARCH_IN_POLICEZONE
						&& !targs.contains(nhP)) {
					if (((Area) nhpE).worldGraphArea.ZoneNumberForPolice == myZoneNumber
							&& ((Area) nhpE).worldGraphArea.policeZoneNumber == myPoliceZone.policeZoneNumber) {
						targs.add(nhP);
					}
				}
				if (searchMode == SearchMode.SEARCH_IN_SUBZONE) {
					if (((Area) nhpE).worldGraphArea.ZoneNumberForPolice == myZoneNumber
							&& ((Area) nhpE).worldGraphArea.subZoneNumberForPolice == mySubZoneNumber
							&& !targs.contains(nhP)) {
						targs.add(nhP);
					}
				}
			}
		}

		log("targs: " + targs);
		if (targs.size() > 0) {
			int target = -1;
			int minDist = Integer.MAX_VALUE;
			for (int a : targs) {
				Entity e = model.getEntityByInt(a);
				int dist = (int) Mathematic.getDistance(me().getX(), me()
						.getY(), ((Area) e).getX(), ((Area) e).getY());
				if (minDist > dist) {
					minDist = dist;
					target = e.getID().getValue();
				}
			}
			clear(target);
		}

		for (Refuge r : myPoliceZone.refuges) {
			Area refuge = (Area) model.getEntity(r.getID());
			if (isInChangeSet(r.getID())
					&& isReachable(r.getID())
					&& (!refuge.isBlockadesDefined() || refuge.worldGraphArea.blockades
							.size() == 0)) {
				haveReachableRefuge = true;
				break;
			}
		}
		if (model.getEntity(me().getPosition()) instanceof Refuge) {
			haveReachableRefuge = true;
		} else if (!haveReachableRefuge && myPoliceZone.refuges.size() > 0) {
			for (Refuge r : myPoliceZone.refuges)
				wg.setTarget(((Area) r).worldGraphArea);
			log("clearing to policezonerefuge ");
			ArrayList<EntityID> path = wg
					.getMinPath(
							((Area) model.getEntity(me().getPosition())).worldGraphArea.enterances,
							false);
			clear(path.get(path.size() - 1).getValue());
		}
	}

	protected void search() throws ActionCommandException {
		log("search###########");
		if (target != null && me().getPosition().equals(target))
			target = null;

		if (target == null) {
			int minTime = Integer.MAX_VALUE;
			for (Building building : modelBuildings)
				if (building.lastTimeVisit < minTime && !building.isOnFire()) {
					minTime = building.lastTimeVisit;
					if (minTime == -2)
						break;
				}
			target = findTarget(minTime);
		}
		clear(target.getValue());
	}

	ArrayList<Integer> myPoliceZoneRefuges = new ArrayList<Integer>();

	protected void decide() throws ActionCommandException {
		log("Changeset: " + changeSet.getChangedEntities());
		log("I HAVE HEARD DDDDDDDDDDDD: : " + policeDecisionHeard1);
		log("My subzone: " + mySubZone + "My policeZone: " + myPoliceZone);
		if (!haveIHeardFromRadar)
			setMyZone(true);

		// for bug of current zoning

		if (mySubZone != null && mySubZone.buildings.size() < 5) {
			mySubZone = zones[0].policeZones[0].subZones[0];
		}
		// server bugs
		if (wg.myEnterances.size() == 0)
			clear(me().getPosition().getValue());
		// set LastTimeVisit
		setLastTimeVisit();

		helpNearStuckAgents(MAX_DIST_FOR_HELP);
		checkHelpIn(SearchMode.SEARCH_IN_POLICEZONE);
		if (mySubZone != null && myZoneNumber != -1)
			fireSearch(mySubZone.buildings);
		noBlockade();
		search();
	}

	int nearAgentPOS = -1;

	public void helpNearStuckAgents(int distance) throws ActionCommandException {
		log("start of help near???????");
		ArrayList<Integer> targets = new ArrayList<Integer>();
		log("CLEARD: " + clearedIDs);
		log("LONGNHP:::" + longnHP);
		log("nearAgentPos1111 " + nearAgentPOS);
		//
		//
		if ((nearAgentPOS != -1
				&& isReachable((model.getEntityByInt(nearAgentPOS)).getID()) && changeSet
				.getChangedEntities().contains(
						model.getEntityByInt(nearAgentPOS).getID()))
				&& (!(((Area) model.getEntityByInt(nearAgentPOS))
						.isBlockadesDefined()) || ((((Area) model
						.getEntityByInt(nearAgentPOS)).isBlockadesDefined()))
						&& ((Area) model.getEntityByInt(nearAgentPOS)).worldGraphArea.blockades
								.size() == 0)) {
			log("1");
			clearedIDs.add(nearAgentPOS);
			nearAgentPOS = -1;
		}
		log("now cleard::: " + clearedIDs);
		// XXX TODO I HAVE TO CHANGE THISSSSSS
		for (int nhpPos : longnHP.values()) {
			if (!clearedIDs.contains(nhpPos)
					&& Mathematic.getDistance(
							((Area) model.getEntityByInt(nhpPos)).getX(),
							((Area) model.getEntityByInt(nhpPos)).getY(),
							((Human) me()).getX(), ((Human) me()).getY()) < MAX_DIST_FOR_HELP) {
				log("3");
				targets.add(nhpPos);
			}
		}
		for (Entity entity : model.getObjectsInRange(((Human) me()).getX(),
				((Human) me()).getY(), MAX_DIST_FOR_HELP))
			if (entity instanceof Refuge
					&& !clearedIDs.contains(entity.getID().getValue())) {
				log("R added: " + entity.getID().getValue());
				targets.add(entity.getID().getValue());
			}

		log("NEAR!: " + targets);

		if (nearAgentPOS == -1) {
			log("4");
			int minDist = Integer.MAX_VALUE;
			for (int a : targets) {
				Entity e = model.getEntityByInt(a);
				int steps = (wg
						.getMinPath(((Area) model.getEntity(((Human) me())
								.getPosition())).worldGraphArea,
								((Area) e).worldGraphArea, false)).size();
				if (minDist > steps) {
					minDist = steps;
					nearAgentPOS = a;
					log("NAP settt: " + nearAgentPOS);
				}
			}
		}
		log("NEARAGENT: " + nearAgentPOS);
		if (nearAgentPOS != -1) {
			log("5");
			clear(nearAgentPOS);
		}
	}

	public void moveToNearRoads(Building target) throws ActionCommandException {
		wg.clearTargets();
		for (Road road : target.visionRoads.keySet()) {
			for (Point point : target.visionRoads.get(road)) {
				if (me().getPosition().getValue() == road.getID().getValue()) {
					ArrayList<EntityID> path = new ArrayList<EntityID>();
					path.add(road.getID());
					log("1");
					clear(path, point.getX(), point.getY());
				}
				road.worldGraphArea.setAreaAsTarget(true);
				break;
			}
		}
		for (Building building : target.visionBuildings) {
			if (me().getPosition().getValue() == building.getID().getValue()) {
				ArrayList<EntityID> path = new ArrayList<EntityID>();
				path.add(building.getID());
				log("LTS:  " + building.lastTimeSeen);
				log("2");
				clear(path, building.getX(), building.getY());
			}
			building.worldGraphArea.setAreaAsTarget(true);
			break;
		}

		int x = -1, y = -1;
		ArrayList<EntityID> path = wg
				.getMinPath(
						((Area) model.getEntity(me().getPosition())).worldGraphArea.enterances,
						false);
		for (Road r : target.visionRoads.keySet())
			if (path.get(path.size() - 1).getValue() == r.getID().getValue()) {
				for (Point point : target.visionRoads.get(r)) {
					x = point.getX();
					y = point.getY();
					log("3:::::: " + path);
					clear(path, x, y);
				}
			}

		for (Building b : target.visionBuildings)
			if (path.get(path.size() - 1).getValue() == b.getID().getValue()) {
				x = b.getX();
				y = b.getY();
				log("4");
				clear(path, x, y);
			}
	}

	// private void clearToFire(ArrayList<Building> allbuildings)
	// throws ActionCommandException {
	// wg.clearTargets();
	// boolean shouldAdd = true;
	// boolean isThereAnyTarget = false;
	// for (Building b : allbuildings) {
	// if (b.isOnFire()) {
	// for (Road r : b.visionRoads.keySet()) {
	// if (((Area) r).worldGraphArea.blockades.size() == 0
	// && isReachable(r.getID()) && r.lastTimeSeen > 0) {
	// shouldAdd = false;
	// break;
	// }
	// }
	// if (shouldAdd) {
	// isThereAnyTarget = true;
	// for (Road r2 : b.visionRoads.keySet()) {
	// ((Area) model.getEntity(r2.getID())).worldGraphArea
	// .setAreaAsTarget(true);
	// }
	// }
	// }
	// }
	// if (isThereAnyTarget) {
	// ArrayList<EntityID> minPath = wg.getMinPath(((Area) model
	// .getEntity(me().getPosition())).worldGraphArea.enterances,
	// false);
	// if (minPath.size() != 0)
	// clear(minPath.get(minPath.size() - 1).getValue());
	// }
	// }

	public void setLastTimeVisit() {
		for (Entity entity : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			Civilian civilian = (Civilian) entity;
			if (model.getEntity(civilian.getPosition()) instanceof Building) {
				Building building = (Building) model.getEntity(civilian
						.getPosition());
				for (Building build : building.nearBuildings)
					if (build.lastTimeVisit == -1)
						build.lastTimeVisit = -2;

			}
		}
		// setting last time visit
		for (EntityID id : changeSet.getChangedEntities()) {
			Entity IDe = model.getEntity(id);
			if (IDe instanceof Building) {
				Building b = (Building) IDe;
				if (!b.hasIntersectWithBuilding(new Line(new Point(me().getX(),
						me().getY()), new Point(b.getX(), b.getY())))) {
					boolean isThereCV = false;
					for (Entity cv : model
							.getEntitiesOfType(StandardEntityURN.CIVILIAN))
						if (((Civilian) cv).getPosition() == b.getID()) {
							isThereCV = true;
							break;
						}

					if (isThereCV && !isReachable(b.getID()))
						continue;
					b.lastTimeVisit = time;
				}
			}
		}

	}

	public EntityID target = null;

	private void fireSearch(ArrayList<Building> buildings)
			throws ActionCommandException {
		log("start fsearch!@@@@@@@@@@");
		String st = " ";
		for (Building b : mySubZone.buildings)
			st += b.getID().getValue();
		log("MYSUBZONENEENEENENEN: " + st);
		boolean shouldIClearRoads = false;
		wg.clearTargets();
		for (Building b : buildings)
			if (b.lastTimeSeen <= -1) {
				shouldIClearRoads = true;
				b.worldGraphArea.setAreaAsTarget(true);
			}

		if (shouldIClearRoads) {
			ArrayList<EntityID> path = wg
					.getMinPath(
							((Area) model.getEntity(me().getPosition())).worldGraphArea.enterances,
							false);
			// TODO khare age size path 0 nabooooood!
			log("my Enterances: " + wg.myEnterances);
			log("path: " + path);
			if (path.size() <= 2)
				clear(path.get(path.size() - 1).getValue());
			moveToNearRoads((Building) model
					.getEntity(path.get(path.size() - 1)));
		}
	}

	public EntityID findTarget(int minVisitTime) {
		wg.clearAreas();

		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		ArrayList<Enterance> start = ((Area) model
				.getEntity(me().getPosition())).worldGraphArea.enterances;

		for (Enterance enterance : start)
			if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer) {
				if (/*
					 * enterance.isItConnectedToNeighbour &&
					 */!enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.area.modelArea.lastTimeVisit <= minVisitTime
							&& (neighbour.area.modelArea instanceof Building || neighbour.area.modelArea instanceof Refuge)
							&& !((Building) neighbour.area.modelArea)
									.isOnFire())
						return neighbour.area.modelArea.getID();

					neighbour.mark = true;
					for (Enterance internal : neighbour.area.enterances) {
						if (!internal.mark) {
							if (internal.area.modelArea.lastTimeVisit <= minVisitTime
									&& (internal.area.modelArea instanceof Building || internal.area.modelArea instanceof Refuge)
									&& !((Building) internal.area.modelArea)
											.isOnFire())
								return internal.area.modelArea.getID();

							internal.mark = true;
							newLayer.add(internal);
						}
					}
				}
			}

			layer = newLayer;
		}

		return null;
	}

	private void clear(ArrayList<EntityID> way, int x, int y)
			throws ActionCommandException {
		Collection<Blockade> blockades = ((Area) model.getEntity(me()
				.getPosition())).worldGraphArea.getBadBlockades(model, me()
				.getX(), me().getY(), 500);
		if (blockades.size() != 0) {
			Blockade blockade = blockades.iterator().next();
			Sclear(blockade);
			return;
		}

		for (int i = 1; i <= 2; i++)
			if (way.size() > i) {
				Entity dest = model.getEntity(way.get(i));
				if (dest instanceof Area) {
					Area area = (Area) dest;
					Collection<Blockade> badBlockades = area.worldGraphArea
							.getBadBlockades(model, me().getX(), me().getY(),
									500);
					for (Blockade badBlockade : badBlockades)
						if (findDistanceTo(badBlockade, me().getX(), me()
								.getY()) < distance) {
							Sclear(badBlockade);
							return;
						}
				}
			}

		int minDistFromBlockade = Integer.MAX_VALUE;
		Blockade targBlockade = null;
		for (Road r : ((Area) model.getEntity(me().getPosition())).nearRoads) {
			Entity roadEntity = ((Area) model.getEntity(r.getID()));
			if (((Area) roadEntity).isBlockadesDefined()) {
				Blockade blockade = null;
				for (EntityID b : ((Area) roadEntity).getBlockades()) {
					blockade = (Blockade) (model.getEntity(b));
					double d = findDistanceTo(blockade, me().getX(), me()
							.getY());
					if (d < distance && isInChangeSet(blockade.getID())) {
						if (d < minDistFromBlockade) {
							minDistFromBlockade = (int) d;
							targBlockade = blockade;
						}
					}
				}
			}
		}

		if (targBlockade != null)
			clear(targBlockade.getID());

		move(way, x, y);
	}

	private void clear(int end) throws ActionCommandException {
		Collection<Blockade> blockades = ((Area) model.getEntity(me()
				.getPosition())).worldGraphArea.getBadBlockades(model, me()
				.getX(), me().getY(), 500);
		if (blockades.size() != 0) {
			Blockade blockade = blockades.iterator().next();
			Sclear(blockade);
			return;
		}

		ArrayList<EntityID> way = wg.getMinPath(me().getPosition(), model
				.getEntityByInt(end).getID(), false);

		log("way bala: " + way);

		for (int i = 1; i <= 2; i++)
			if (way.size() > i) {
				Entity dest = model.getEntity(way.get(i));
				if (dest instanceof Area) {
					Area area = (Area) dest;
					Collection<Blockade> badBlockades = area.worldGraphArea
							.getBadBlockades(model, me().getX(), me().getY(),
									500);
					for (Blockade badBlockade : badBlockades)
						if (findDistanceTo(badBlockade, me().getX(), me()
								.getY()) < distance) {
							Sclear(badBlockade);
							return;
						}
				}
			}

		int minDistFromBlockade = Integer.MAX_VALUE;
		Blockade targBlockade = null;
		for (Road r : ((Area) model.getEntity(me().getPosition())).nearRoads) {
			Entity roadEntity = ((Area) model.getEntity(r.getID()));
			if (((Area) roadEntity).isBlockadesDefined()) {
				Blockade blockade = null;
				for (EntityID b : ((Area) roadEntity).getBlockades()) {
					blockade = (Blockade) (model.getEntity(b));
					double d = findDistanceTo(blockade, me().getX(), me()
							.getY());
					if (d < distance / 4 && isInChangeSet(blockade.getID())) {
						if (d < minDistFromBlockade) {
							minDistFromBlockade = (int) d;
							targBlockade = blockade;
						}
					}
				}
			}
		}

		if (targBlockade != null)
			clear(targBlockade.getID());
		log("way here: " + way);
		move(way);
	}

	private void Sclear(Blockade b) throws ActionCommandException {
		ArrayList<EntityID> path = wg
				.getMinPath(
						wg.myEnterances,
						((Area) model.getEntity(b.getPosition())).worldGraphArea,
						false);
		if (path.size() == 0)
			path.add(me().getPosition());
		double dist = model.getDistance(me(), b);
		for (int i = 2; i < b.getApexes().length; i += 2) {
			Point2D po = new Point2D(
					(b.getApexes()[i - 2] + b.getApexes()[i]) / 2,
					(b.getApexes()[i - 1] + b.getApexes()[i + 1]) / 2);
			double tmpDist = GeometryTools2D.getDistance(new Point2D(me()
					.getX(), me().getY()), po);
			if (tmpDist < dist) {
				dist = tmpDist;
			}
		}
		Point2D po = new Point2D(
				(b.getApexes()[0] + b.getApexes()[b.getApexes().length - 2]) / 2,
				(b.getApexes()[1] + b.getApexes()[b.getApexes().length - 1]) / 2);
		double tmpDist = GeometryTools2D.getDistance(new Point2D(me().getX(),
				me().getY()), po);
		if (tmpDist < dist) {
			dist = tmpDist;
		}

		if (dist < distance && isInChangeSet(b.getID())) {
			// log("Clear");
			clear(b.getID());
		} else {
			if (path.size() > 2) {
				ArrayList<EntityID> path2 = new ArrayList<EntityID>();
				path2.add(path.get(0));
				path2.add(path.get(1));
				log("Path2: " + path2);
				move(path2);
			} else
				move(path, b.getX(), b.getY());
		}
	}

	private int findDistanceTo(Blockade b, int x, int y) {
		List<Line2D> lines = GeometryTools2D.pointsToLines(
				GeometryTools2D.vertexArrayToPoints(b.getApexes()), true);
		double best = Double.MAX_VALUE;
		Point2D origin = new Point2D(x, y);
		for (Line2D next : lines) {
			Point2D closest = GeometryTools2D.getClosestPointOnSegment(next,
					origin);
			double d = GeometryTools2D.getDistance(origin, closest);
			if (d < best)
				best = d;
		}
		return (int) best;
	}

	public EntityID findTargetForNoBlockade(int minVisitTime) {
		wg.clearAreas();

		ArrayList<Enterance> layer = new ArrayList<Enterance>();
		ArrayList<Enterance> start = wg.myEnterances;

		for (Enterance enterance : start)
			if (!enterance.mark) {
				enterance.mark = true;
				layer.add(enterance);
			}

		while (layer.size() > 0) {
			ArrayList<Enterance> newLayer = new ArrayList<Enterance>();
			for (Enterance enterance : layer) {
				if (!enterance.neighbour.mark) {
					Enterance neighbour = enterance.neighbour;
					if (neighbour.star
							&& neighbour.area.modelArea.lastTimeVisit <= minVisitTime
							&& (neighbour.area.modelArea instanceof Building || neighbour.area.modelArea instanceof Refuge)
							&& !((Building) neighbour.area.modelArea)
									.isOnFire())
						return neighbour.area.modelArea.getID();

					neighbour.mark = true;
					for (Enterance internal : neighbour.area.enterances) {
						if (!internal.mark) {
							if (internal.star
									&& internal.area.modelArea.lastTimeVisit <= minVisitTime
									&& (internal.area.modelArea instanceof Building || internal.area.modelArea instanceof Refuge)
									&& !((Building) internal.area.modelArea)
											.isOnFire())
								return internal.area.modelArea.getID();

							internal.mark = true;
							newLayer.add(internal);
						}
					}
				}
			}

			layer = newLayer;
		}
		return null;
	}

	public void noBlockade() throws ActionCommandException {
		if (mySubZone == null)
			return;

		for (Building building : modelBuildings)
			building.worldGraphArea.clearStar();

		for (Building building : mySubZone.buildings)
			building.worldGraphArea.setStar();
		int minTime = Integer.MAX_VALUE;
		for (Building building : mySubZone.buildings)
			if (building.isOnFire() == false
					&& building.lastTimeVisit < minTime) {
				minTime = building.lastTimeVisit;
				if (minTime == -2)
					break;
			}

		EntityID target = findTargetForNoBlockade(minTime);
		if (target != null)
			clear(target.getValue());
	}

}
