package geneticAlgorithm;

import geometry.Mathematic;

import java.util.ArrayList;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardWorldModel;

public class State {
	private ArrayList<Civilian> cvs = null;
	private AmbulanceTeam[] list = null;
	private int fitness = -1;
	private StandardWorldModel model = null;
	public ArrayList<Human> agents = null;
	public ArrayList<AmbulanceTeam> ambs = null;

	public State(ArrayList<Civilian> c, ArrayList<Human> a,
			ArrayList<AmbulanceTeam> ambulances, StandardWorldModel m) {
		cvs = c;
		agents = a;
		ambs = ambulances;
		model = m;
	}

	public int getFitness() {
		return fitness;
	}

	public AmbulanceTeam[] getList() {
		return list;
	}

	public ArrayList<Civilian> getCvs() {
		return cvs;
	}

	public ArrayList<Human> getAgents() {
		return agents;
	}

	public void generateList() {
		if (cvs.size() == 0 && agents.size() == 0)
			return;
		list = new AmbulanceTeam[cvs.size() + agents.size()];
		for (AmbulanceTeam ambulanceTeam : ambs)
			ambulanceTeam.freetime = ambulanceTeam.firstFreeTime;
		int i = 0;
		for (Human a : agents) {
			double minDist = Integer.MAX_VALUE, d = Integer.MAX_VALUE;
			AmbulanceTeam nearestAmb = null;
			for (AmbulanceTeam amb : ambs) {
				d = Mathematic.getDistance(a.getX(), a.getY(), ((Area) model
						.getEntity(amb.curPos)).getX(), ((Area) model
						.getEntity(amb.curPos)).getY());
				if (amb.freetime < a.deadtime && d < minDist) {
					minDist = d;
					nearestAmb = amb;
				}
			}
			list[i] = nearestAmb;
			if (list[i] != null) {
				nearestAmb.freetime += a.getBuriedness() + 10;
				nearestAmb.curPos = a.getPosition();
			}
			i++;
		}
		for (Civilian civilian : cvs) {
			double minDist = Integer.MAX_VALUE, d = Integer.MAX_VALUE;
			AmbulanceTeam nearestAmb = null;
			for (AmbulanceTeam amb : ambs) {
				d = Mathematic.getDistance(civilian.getX(), civilian.getY(),
						((Area) model.getEntity(amb.curPos)).getX(),
						((Area) model.getEntity(amb.curPos)).getY());
				if (amb.freetime < civilian.deadtime && d < minDist) {
					minDist = d;
					nearestAmb = amb;
				}
			}
			list[i] = nearestAmb;
			if (list[i] != null) {
				nearestAmb.freetime += civilian.getBuriedness() + 20;
				nearestAmb.curPos = civilian.nearestRefuge;
			}
			i++;
		}
	}

	public double getDistance(Area a1, Area a2) {
		double d = Math.hypot(a1.getX() - a2.getX(), a2.getY() - a2.getY());
		return d;
	}
}
